"""*VPRM* : Vegetation Photosynthesis and Respiration Model.


The VPRM model is a parametrized model that estimates
the photosynthesis and respiration of vegetation based on satellite observations.
The model was developed by [Mahadevan_2008]_

"""
from __future__ import annotations
import logging
from typing import Union
import numpy as np
import pandas as pd
import xarray as xr

bandType = Union[np.ndarray,xr.DataArray]


def calculate_vegetation_indices(
    nir: bandType,
    swir: bandType,
    red: bandType,
    blue: bandType,
    # EVI parameters
    vprm_g: float = 2.5,
    vprm_c1: float = 6.0,
    vprm_c2: float = 7.5,
    vprm_l: float = 1.0,
    # Clipping 
    clip_evi: bool = False,
):
    """Calculate the vrpm products from the satellite observations.

    The formulas are the following:

    .. math::

        \\mathrm{EVI} &= \\frac{G \\cdot (\\mathrm{NIR} - \\mathrm{RED})}{(\\mathrm{NIR} + C_1 \\cdot \\mathrm{RED} - C_2 \\cdot \\mathrm{BLUE}) + L}
        
        \\newline

        \\mathrm{LSWI} &= \\frac{\\mathrm{NIR} - \\mathrm{SWIR}}{\\mathrm{NIR} + \\mathrm{SWIR}}
        
        \\newline

        \\mathrm{NDVI} &= \\frac{\\mathrm{NIR} - \\mathrm{RED}}{\\mathrm{NIR} + \\mathrm{RED}}


    
    The input bands can by numpy arrays or xarray DataArrays.

    :param nir: Near Infrared band
    :param swir: Shortwave Infrared band
    :param red: Red band
    :param blue: Blue band

    :param vprm_g: Gain factor for EVI
    :param vprm_c1: Coefficient 1 for EVI
    :param vprm_c2: Coefficient 2 for EVI
    :param vprm_l: Coefficient L for EVI

    :param clip_evi: Clip the EVI values between 0 and 1. 
        As the equation for EVI does not produce a proper index,
        values can be negative or above 1 if not clipped.
    :return: Tuple with the EVI, LSWI and NDVI
    
    """
    evi = vprm_g * (nir - red) / (nir + vprm_c1 * red - vprm_c2 * blue + vprm_l)
    lswi = (nir - swir) / (nir + swir)
    ndvi = (nir - red) / (nir + red)

    if clip_evi:
        evi = np.clip(evi, 0, 1)

    return evi, lswi, ndvi



def calculate_vprm_emissions(df: pd.DataFrame, df_vprm: pd.DataFrame) -> pd.DataFrame:
    """Calculate the emissions using the VPRM model.

    This function uses timeseries of vegetation indices, temperature and radiation 
    to calculate the
    respiration and photosynthesis emissions of vegetation.

    It handles various vegetation types.
    Each vegetation type has its own parameters for the VPRM model.

    It also includes extensions of the VPRM model for urban areas.


    The equations used are the following:

    PAR (Photosynthetically Active Radiation) is calculated from the shortwave radiation:
    
    .. math::
        \\mathrm{PAR} = \\frac{\\mathrm{RAD}}{0.505}
    
    Respiration is calculated from the temperature:
    
    .. math::
        \\mathrm{Resp} = \\alpha * T + \\beta
    
    The Gross Ecosystem Exchange (GEE) is calculated from the temperature, PAR and the vegetation indices:
    
    .. math::
        \\mathrm{GEE} = \\lambda * T_{scale} * P_{scale} * W_{scale} * \\mathrm{EVI} * \\frac{ \\mathrm{PAR} }{1 + \\frac{\\mathrm{PAR}}{PAR0}}

    where the different scales are:

    - :math:`T_{scale}`: Temperature scale

    .. math::
        T_{\\text{scale}} = \\frac{(T - T_{\\text{min}}) \\cdot (T - T_{\\text{max}})}{(T - T_{\\text{min}}) \\cdot (T - T_{\\text{max}}) + (T - T_{\\text{opt}})^2} \\text{if } T \\geq T_{\\text{min}} \\text{ else } 0
        


    - :math:`P_{scale}`: Photosynthesis scale

    .. math::
        P_{scale} = \\frac{1 + \\mathrm{LSWI}}{2}
    
    - :math:`W_{scale}`: Water scale

    .. math::
        W_{scale} = \\frac{1 + \\mathrm{LSWI}}{1 + \\mathrm{LSWI}_{max}}

    The Net Ecosystem Exchange (NEE) is calculated from the respiration and GEE.
    
    .. math::
        \\mathrm{NEE} = \\mathrm{Resp} + \\mathrm{GEE}

        
    Units for all fluxes (NEE, GEE, Resp, ...) are

    .. math:: 
        \\frac{\\mu mol_{\\mathrm{CO2}}}{m^2 * s}

        
    Urban modifications

    The VPRM model can be extended to urban areas according to  [Urban_VPRM]_ .
    
    - A "urban temperature" is used instead of the global temperature to represent
        the urban heat island phenomenon.
    - The formula for :math:`P_{scale}` is modified to

    .. math::
        P_{scale} = \\frac{\\mathrm{EVI} - \\mathrm{EVI}_{min}}{\\mathrm{EVI}_{max} - \\mathrm{EVI}_{min}}

    - The respiration is calculated differently
    
    .. math::
        \\mathrm{Resp} = \\frac{\\mathrm{Resp_{e-init}}}{2} * (1 - \\mathrm{ISA}) + \\frac{\\mathrm{EVI} + \\mathrm{EVI}_{min} * \\mathrm{ISA}}{\\mathrm{EVI}_{ref}} * \\frac{\\mathrm{Resp_{e-init}}}{2}
    
    where :math:`\\mathrm{Resp_{e-init}}` is the basic vprm respiration and :math:`\\mathrm{ISA}` is the impervious surface area at the vegetation location.
    
    .. warning::
        The urban VPRM model is currently not fully implemented.
    
    
    :param df: Dataframe with the observations. It must be a multiindex dataframe with the following columns:
    
        - `RAD`: Shortwave radiation in W/m2
        - ('T', 'global'): Temperature in degC
        - (vegetation_type, 'lswi'): Land Surface Water Index 
        - (vegetation_type, 'evi'): Enhanced Vegetation Index
        - ('T', 'urban'): Optional for urban VPRM. Temperature in degC (urban area)
    
    :param df_vprm: Dataframe with the VPRM parameters.
        Each row must correspond to a vegetation type and have the following columns:
        
        - `alpha`: Respiration parameter
        - `beta`: Respiration parameter
        - `lambda`: Photosynthesis parameter
        - `Tmin`: Minimum temperature for photosynthesis
        - `Topt`: Optimal temperature for photosynthesis
        - `Tmax`: Maximum temperature for photosynthesis
        - `Tlow`: Low temperature for photosynthesis
        - `PAR0`: Photosynthetically Active Radiation parameter
        - `is_urban`: Boolean indicating if the vegetation type is urban (optional, default is False)
   
    :return: Dataframe with the emissions. Some columns are added

        - (vegetation_type, 'resp_min'): Respiration at the minimum temperature
        - (vegetation_type, 'resp_max'): Respiration at the maximum temperature
        - (vegetation_type, 'resp'): Respiration
        - (vegetation_type, 'gee'): Gross Ecosystem Exchange
        - (vegetation_type, 'nee'): Net Ecosystem Exchange (nee = gee - resp)
    """
    logger = logging.getLogger(__name__)
    df = df.copy()

    df_vprm['resp_min'] = df_vprm['alpha'] * df_vprm["Tlow"] + df_vprm['beta']

    # Photosynthetically Active Radiation (PAR, μmol m−2 s−1) 
    # Conversion from orginal vprm paper, assuming RAD is shortwave radiation
    df['PAR'] = df['RAD'] / 0.505


    if 'is_urban' not in df_vprm.columns:
        df_vprm['is_urban'] = False
    if any(df_vprm['is_urban']):
        # Ensure that the urban temperature is present
        if ('T', 'urban') not in df.columns:
            raise ValueError("Urban VPRM is activated but the urban temperature is missing in the dataframe")

    for vegetation_type in df_vprm.index:
        if not all([(vegetation_type, index) in df.columns for index in ['lswi', 'evi']]):
            logger.warning(f"Missing {vegetation_type} in the observation dataframe, skipping")
            continue


        # Add to the metot the paramters from the satellite observations 
        # Use interpolation to get the values for the missing dates
        lswi = df[(vegetation_type, 'lswi')]
        evi = df[(vegetation_type, 'evi')]

        
            
        is_urban = df_vprm.loc[vegetation_type, 'is_urban']

        Tmin = df_vprm.loc[vegetation_type, 'Tmin']
        Topt = df_vprm.loc[vegetation_type, 'Topt']
        Tmax = df_vprm.loc[vegetation_type, 'Tmax']
        Tlow = df_vprm.loc[vegetation_type, 'Tlow']

        # Get correct temperature df
        if is_urban:
            temperature = df[('T', 'urban')]
        else:
            temperature = df[('T', 'global')]
        
        # Calculate the respiration 

        # Resp = alpha * T + beta
        resp  = df_vprm.loc[vegetation_type, 'alpha'] * temperature + df_vprm.loc[vegetation_type, 'beta']

        # Under t low, use a contsant value 
        mask_low_T = temperature <= Tlow
        resp.loc[mask_low_T] =  df_vprm.loc[vegetation_type, 'resp_min']



        ## Split the urban vegetation into two parts
        ## initial ecosystem respiration (authotropphic + heterotropohic)
        #df[(vegetation_type, 'resp_e_init')] = df[(vegetation_type, 'resp _urban')] 
    #
        ## Heterotrophic respiration
        ## isa = impervious surface areas 
        #isa = 0.5 # ??? not  sure what value this should be
        #r  = (1- isa) * df[(vegetation_type, 'resp_e_init')] / 2.
    #
        ## Get reference  the yearly minimum of EVI at a reference Forest 
        ## (representing leaf-off, woody biomass autotrophic respiration
        #evi_ref = df_means[(vegetation_type_ref[vegetation_type], 'EVI')]
        #r_a = (evi + np.nanmin(evi_ref) * isa) / evi_ref * resp_e_init / 2.
        #
        ## Bring the two components together
        #resp  = r  + r_a



        # GEE
        Tscale  = (temperature - Tmin) * (temperature- Tmax)
        Tscale  = Tscale  / (Tscale  - (temperature - Topt) ** 2)
        Tscale [temperature <= Tmin] = 0.0
        df[(vegetation_type, 'Tscale')] = Tscale 


        # Typical summer values for LSWI, EVI and Wscale
        # To be replaced with satellite based parameters
        Wscale  = (1 + lswi) / (1 + np.nanmax(lswi))
        df[(vegetation_type, 'Wscale')] = Wscale

        if is_urban:
            # Simpler EVI forumalation in urban VPRM
            Pscale  = (evi - np.nanmin(evi)) / (np.nanmax(evi) - np.nanmin(evi))
        else:
            Pscale  = (1 + lswi) / 2.0
        df[(vegetation_type, 'Pscale')] = Pscale

        gee  = (
            df_vprm.loc[vegetation_type, "lambda"]
            * Tscale 
            * Pscale 
            * Wscale 
            * evi
            * df['PAR']
            / (1 + df['PAR'] / df_vprm.loc[vegetation_type, "PAR0"])
        )


        #  VPRM produces umoles/m2/s
        df[(vegetation_type, 'resp')] = resp   
        df[(vegetation_type, 'gee')] = gee   
        df[(vegetation_type, 'nee')] = resp  + gee   
    
    return df



"""Function for doing speciation.

The function speciate_inventory() is the main function.
Other functions are hardcoded for several substances.
"""

from __future__ import annotations

from os import PathLike
from typing import TYPE_CHECKING, Any

import pandas as pd
import xarray as xr
import numpy as np

from emiproc.utilities import get_country_mask

if TYPE_CHECKING:
    from emiproc.inventories import Category, CatSub, Inventory


def read_speciation_table(
    path: PathLike, drop_zeros: bool = False, check_sum: bool = True, **kwargs
) -> xr.DataArray:
    """Read a speciation table from a file.

    Format of the file:

    ```
    # Any comment line starting with # is ignored
    # The first line is the header, it contains optional information
    # to specify speciation factors for each category.
    category,country,substance0,substance1,substance2, ...
    cat0,c0,0.5,0.2,0.1, ...
    cat1,c0,0.3,0.5,0.0, ...
    cat0,c1,0.2,0.3,0.5, ...
    ...
    ```


    The table can contain optional dimensions:
        - country : must follow the ISO3 naming convention
            In case cells belong to no country and have emissions, you will
            need to set a default value for the speciation ratios.
            This can be done by adding a row with the country set to `-99`.
        - category
        - type : Whether applies to 'gridded' or 'shapped'
        - year

    Ratios are the fraction of the weight of the speciated substance.
    As emissions are given in mass/unit of time, the sum of the ratios
    for a given speciiation must be 1.

    :arg path: The path of the speciation table.
    :arg drop_zeros: Whether to drop the speciation ratios that sum to 0.
    :arg check_consistency: Whether to check that the speciation ratios sum to 1.
        If this is set to False, the check is not performed.
    :arg kwargs: Additional arguments to pass to :py:func:`pandas.read_csv`.

    :returns: The speciation ratios.
        The speciation ratios are the weight fraction conversion from the
        speciated substance.

    """
    df = pd.read_csv(path, comment="#", **kwargs)

    # Reserved columns
    columns_types = {
        "category": str,
        "country": str,
        "type": str,
        "year": int,
    }

    new_substances = [col for col in df.columns if col not in columns_types.keys()]

    # Get the ratios data
    ratios = df[new_substances].to_numpy()

    da = xr.DataArray(
        ratios,
        dims=["speciation", "substance"],
        coords={
            "speciation": df.index,
            "substance": new_substances,
        }
        | {
            # Add the optional dimensions that describe the speciation
            k: ("speciation", df[k].to_numpy().astype(t))
            for k, t in columns_types.items()
            if k in df.columns
        },
    )

    if drop_zeros:
        mask_zero = da.sum("substance") == 0
        da = da[~mask_zero]

    # Check that all the speciation ratios sum to 1
    mask_close = np.isclose(da.sum("substance"), 1.0)
    if check_sum and not mask_close.all():
        raise ValueError(
            "The speciation ratios must sum to 1, but the following rows don't:"
            f" {da[~mask_close]}"
            "You can set `check_sum=False` to ignore this check."
        )

    return da


def ratios_of_category(
    speciation_ratios: xr.DataArray,
    category: Category,
) -> xr.DataArray:
    """Get the speciation ratios for a category.

    :arg speciation_ratios: The speciation ratios.
    :arg category: The category to get the speciation ratios for.

    :returns: The speciation ratios for the category.
    """
    if "category" not in speciation_ratios.coords:
        return speciation_ratios

    mask = speciation_ratios["category"] == category
    return speciation_ratios.loc[dict(speciation=mask)]


def speciate(
    inv: Inventory,
    substance: str,
    speciation_ratios: xr.DataArray,
    drop: bool = True,
    country_mask_kwargs: dict[str, Any] = {},
) -> Inventory:
    """Speciate a substance in an inventory.

    .. note:: Profiles (vertical and temporal) are not speciated.
        The profiles of the speciated substance are simply applied.
        If you have specific profiles for the speciated substance,
        you need to set them after the speciation operation.

    :arg inv: The inventory to speciate.
    :arg substance: The substance to speciate.
    :arg speciation_ratios: The speciation ratios.
        See :py:func:`read_speciation_table` to load them from a file.
        The ratios must be a 2-D data array. One dimension is the substances
        to create with the speciation, the other dimension is called 'speciation'
        and each element is a set of ratios that sum to 1.
        Each speciation can be specific to a category, a country, a year, or a type.
        This is done by adding coordinates on the 'speciation' dimension.

        .. code-block:: python

            <xarray.DataArray (speciation: 5, substance: 2)>
            # Speciation ratios
            array([[0.5, 0.5],
                [0.2, 0.8],
                [0.3, 0.7],
                [0.4, 0.6],
                [0.1, 0.9]])
            Coordinates:
            * speciation  (speciation) int64 0 1 2 3 4
            * substance   (substance) <U7 'CO2_ANT' 'CO2_BIO'
            # Optional dimensions which tell which combination of parameters the ratios apply to
                category    (speciation) <U4 'adf' 'liku' 'adf' 'liku' 'blek'
                type        (speciation) <U7 'gridded' 'gridded' 'gridded' 'shapped' 'shapped'
                year        (speciation) int64 2010 2010 2010 2010 2010


    :arg drop: Whether to drop the speciated substance.
    :arg country_mask_kwargs: If the speciation ratios depend on the country,
        this function is used to pass optional arguments to
        :py:func:`emiproc.utilities.get_country_mask`.


    """
    new_inv = inv.copy()

    # If speciation is not given it is okay
    for dim in ["speciation", "substance"]:
        if dim not in speciation_ratios.coords:
            raise ValueError(
                f"The speciation ratios must have a '{dim}' dimension."
                f" {speciation_ratios.coords=}"
            )

    # Check that it is 2D
    if len(speciation_ratios.dims) != 2:
        raise ValueError(
            f"The speciation ratios must be a 2D array, not {speciation_ratios.dims=}"
        )

    if "year" in speciation_ratios.coords:
        if inv.year is None:
            raise ValueError(
                f"The inventory {inv} does not have a year, but the speciation ratios"
                f" {speciation_ratios} do."
            )
        # Get the speciation ratios for the year of the inventory
        mask_year = speciation_ratios["year"] == inv.year
        speciation_ratios = speciation_ratios.loc[mask_year]

    if "country" in speciation_ratios.coords:
        countries_fractions: xr.DataArray = get_country_mask(
            new_inv.grid, return_fractions=True, **country_mask_kwargs
        )
        # Fill the fraction so that the sum is 1
        # Ensure that a speciation defined in a cell only on some country parts will be
        # applied to the whole cell by filling the missing values with the countries
        countries_fractions = countries_fractions / countries_fractions.sum("country")
        countries_fractions = countries_fractions.fillna(0.0)

    for cat, sub in inv._gdf_columns:
        if sub != substance:
            continue
        # Check that the speciation ratios are defined for this category
        da_ratios = ratios_of_category(speciation_ratios, cat)

        if "type" in speciation_ratios.coords:
            da_ratios = da_ratios.loc[da_ratios["type"] == "gridded"]

        if da_ratios["speciation"].size == 0:
            raise ValueError(
                f"The speciation ratios for {cat} and {substance} is not defined."
            )

        if "country" in speciation_ratios.coords:
            # Calculate the ratio at each cell
            # Make the dim on country instead of speciation
            da_ratios_country = da_ratios.set_index(speciation="country").rename(
                {"speciation": "country"}
            )
            da_ratios_cells = countries_fractions.dot(
                da_ratios_country, dims=["country"]
            )
            # First check that where the sum is 0, the total emissions are 0
            mask_zero_ratios = da_ratios_cells.sum("substance") == 0
            mask_zero_emissions = inv.gdf[(cat, substance)] == 0
            mask_problem = mask_zero_ratios & ~mask_zero_emissions
            missing_value = (
                None
                if "-99" not in da_ratios_country.coords["country"].values
                else da_ratios_country.sel(country="-99")
            )
            if mask_problem.any():
                if missing_value is None:
                    raise ValueError(
                        f"The speciation ratios for {cat} and {substance} is not"
                        " defined for the following cells that have emission values :"
                        f" {da_ratios_cells[mask_problem]['cell']}\n"
                        "You can add a row with country set to -99 to define the"
                        " default speciation ratios for those homeless cells."
                    )
                # Set the missing value
                da_ratios_cells = da_ratios_cells.where(
                    ~mask_problem, other=missing_value
                )

            da_ratios = da_ratios_cells

        else:
            # No country, just speciate every cell the same way
            if da_ratios["speciation"].size > 1:
                raise ValueError(
                    f"The speciation ratios for {cat} and {substance} is not unique:"
                    f" {da_ratios=}."
                )
        # Speciate the gdf
        for new_sub in da_ratios["substance"].values:
            # Simply get the single value
            ratios = da_ratios.sel(substance=new_sub).values
            # Should not happen, but just in case
            if (cat, new_sub) in inv._gdf_columns:
                raise KeyError(f"{cat}/{new_sub} already in the gdf of {inv}")
            new_inv.gdf[(cat, new_sub)] = inv.gdf[(cat, substance)] * ratios
        if drop:
            # Drop the speciated substance
            new_inv.gdf.drop(columns=[(cat, substance)], inplace=True)

    # Now for the gdfs
    for cat in inv.gdfs.keys():
        if substance not in inv.gdfs[cat].columns:
            continue
            # Check that the speciation ratios are defined for this category
        da_ratios = ratios_of_category(speciation_ratios, cat)

        if "type" in speciation_ratios.coords:
            da_ratios = da_ratios.loc[da_ratios["type"] == "shapped"]
        if "country" in speciation_ratios.coords:
            country_mask = get_country_mask(
                inv.gdfs[cat].geometry, **country_mask_kwargs
            )
            mask_no_country = country_mask == "-99"
            if any(mask_no_country) and "-99" not in da_ratios.coords["country"]:
                raise ValueError(
                    f"The speciation ratios for {cat} and {substance} is not defined"
                    f" for the following shapes {inv.gdfs[cat].loc[mask_no_country]}"
                    "You can add a country with iso3='-99' to define the"
                    " default speciation ratios for those homeless shapes."
                )
            da_ratios_country = da_ratios.set_index(speciation="country").rename(
                {"speciation": "country"}
            )
            da_ratios = da_ratios_country.loc[country_mask]
        else:
            if da_ratios["speciation"].size == 0:
                raise ValueError(
                    f"The speciation ratios for {cat} and {substance} is not defined"
                    " for shapped emissions."
                )
            if da_ratios["speciation"].size > 1:
                raise ValueError(
                    f"The speciation ratios for {cat} and {substance} is not unique"
                    f" {da_ratios=}."
                )

        # Speciate the gdf
        for new_sub in da_ratios["substance"].values:
            speciation_ratio = da_ratios.sel(substance=new_sub).values
            new_inv.gdfs[cat][new_sub] = inv.gdfs[cat][substance] * speciation_ratio
        if drop:
            new_inv.gdfs[cat].drop(columns=[substance], inplace=True)

    # profiles
    # this is easy because we can simply duplicate the indexes in substance
    for indexes_name in ["t_profiles_indexes", "v_profiles_indexes"]:
        if not hasattr(inv, indexes_name):
            continue
        indexes: xr.DataArray = getattr(inv, indexes_name)
        if indexes is None or "substance" not in getattr(inv, indexes_name).dims:
            continue

        speciated_indexes = (
            indexes.sel(substance=substance)
            .drop_vars("substance")
            .expand_dims(dim={"substance": da_ratios["substance"].values})
        )
        new_indexes = xr.concat([indexes, speciated_indexes], dim="substance")
        if drop:
            # Remove the substance from the substance coordinate
            new_indexes = new_indexes.drop_sel(substance=[substance])

        setattr(new_inv, indexes_name, new_indexes)

    new_inv.history.append(f"Speciated {substance} to {da_ratios['substance'].values}.")

    return new_inv


def speciate_inventory(
    inv: Inventory,
    speciation_dict: dict[CatSub, dict[CatSub, float]],
    drop: bool = True,
) -> Inventory:
    """Speciate an inventory.

    Speciation is splitting one substance into several substances.
    For example NOx can be split into NO and NO2.
    Replaces the current category/substance with the new one.

    :arg inv: The inventory to speciate.
    :arg speciation_dict: A dict with the speciation rules.
        The keys are the category/substance to speciate.
        The values is a dict mapping new category/substance and the fraction
        of the orignal substance to be specified.
        Ratio is the weight fraction conversion from the speciated substance.
        Note that the ratio don't need to sum to 1, depending on the
        chemical parameters.
    :arg drop: Whether to drop the speciated category/substance.

    :returns: The speciated inventory.
    """
    new_inv = inv.copy()

    for cat_sub, new_species in speciation_dict.items():
        cat, sub = cat_sub
        # Check the there is a substance to speciate
        if cat_sub not in inv._gdf_columns and (
            cat not in inv.gdfs or sub not in inv.gdfs[cat]
        ):
            raise KeyError(f"Cannot speciate: {cat_sub} not in {inv}")

        # Speciate the gdf
        if cat_sub in inv._gdf_columns:
            for new_cat_sub, speciation_ratio in new_species.items():
                # if the new cat/sub is already in the gdf raise an error
                if new_cat_sub in inv._gdf_columns:
                    raise KeyError(f"{new_cat_sub} already in the gdf of {inv}")
                new_inv.gdf[new_cat_sub] = inv.gdf[cat_sub] * speciation_ratio
            if drop:
                new_inv.gdf.drop(columns=cat_sub, inplace=True)
        # Speciate the gdfs
        if cat in inv.gdfs and sub in inv.gdfs[cat].columns:
            for new_cat_sub, speciation_ratio in new_species.items():
                new_cat, new_sub = new_cat_sub
                # if the new cat/sub is already in the gdf raise an error
                if new_cat in inv.gdfs and new_sub in inv.gdfs[new_cat].columns:
                    raise KeyError(
                        f"Cannot speciate: {new_cat_sub} already in the gdfs of {inv}"
                    )
                new_inv.gdfs[new_cat][new_sub] = inv.gdfs[cat][sub] * speciation_ratio
            if drop:
                new_inv.gdfs[cat].drop(columns=sub, inplace=True)

    # Profiles should also be speciated, simply apply the profile to all compounds
    for indexes_name in ["t_profiles_indexes", "v_profiles_indexes"]:
        if not hasattr(inv, indexes_name):
            continue
        indexes: xr.DataArray = getattr(inv, indexes_name)
        if indexes is None or "substance" not in indexes.dims:
            continue
        new_data_arrays = [indexes]

        for cat_sub, new_species in speciation_dict.items():
            cat, sub = cat_sub
            new_species = [sub for cat, sub in new_species]
            speciated_indexes = (
                indexes.sel(substance=sub)
                .drop_vars(["substance"])
                .expand_dims(dim={"substance": [new_sub for new_sub in new_species]})
            )
            new_data_arrays.append(speciated_indexes)
        new_indexes = xr.concat(new_data_arrays, dim="substance")
        # Drop duplicates along the substance dimension
        new_indexes = new_indexes.drop_duplicates("substance")
        if drop:
            # Remove the substance from the substance coordinate
            new_indexes = new_indexes.drop_sel(
                substance=[sub for cat, sub in speciation_dict.keys()]
            )
        setattr(new_inv, indexes_name, new_indexes)

    new_inv.history.append(f"Speciated with {speciation_dict}.")

    return new_inv


def speciate_nox(
    inv: Inventory,
    NOX_TO_NO2: float | dict[Category, float] = 0.18,
    drop: bool = True,
) -> Inventory:
    """Speciate NOx into NO and NO2.

    :arg inv: The inventory to speciate.
    :arg drop: Whether to drop the speciated category/substance.
    :arg NOX_TO_NO2: The fraction of NOx that is speciated to NO2.
        It is possible to use a dict with a different fraction for each category.

        .. note::

            Depending on the sector, this value can vary.

            For most emission sources, the fraction of NO is closer to 95%,
            only for traffic a fraction of 82% may be applied.
            The reason is that oxidation catalysts in diesel engines partly
            oxidize NO to NO2 before it is emitted through the tailpipe.
            In the first decade of 2000,
            the fraction of NO gradually decreased from 95% to about 80%.

            See more:

            * https://www.empa.ch/documents/56101/246436/Trend+NO2+Immissionen+Stadt+2022/ddba8b88-c599-4ed4-8b94-cc24670be683
            * https://www.zh.ch/de/umwelt-tiere/luft-strahlung/luftschadstoffquellen/emissionen-verkehr/abgasmessungen-rsd.html



    :returns: The speciated inventory.

    """
    MOLAR_MASS_NO2 = 46.0
    MOLAR_MASS_NO = 30.0
    MM_RATIO = MOLAR_MASS_NO / MOLAR_MASS_NO2

    if isinstance(NOX_TO_NO2, dict):
        speciation_dict = {
            (cat, "NOx"): {
                (cat, "NO"): (1.0 - ratio) * MM_RATIO,
                (cat, "NO2"): ratio,
            }
            for cat, ratio in NOX_TO_NO2.items()
        }
    elif isinstance(NOX_TO_NO2, float):
        # Make sure the fraction is between 0 and 1
        if NOX_TO_NO2 < 0 or NOX_TO_NO2 > 1:
            raise ValueError(f"NOX_TO_NO2 must be between 0 and 1, not {NOX_TO_NO2}.")
        # Aplly the same fraction to all categories
        speciation_dict = {
            (cat, "NOx"): {
                (cat, "NO"): (1.0 - NOX_TO_NO2) * MM_RATIO,
                (cat, "NO2"): NOX_TO_NO2,
            }
            for cat in inv.categories
            if (cat, "NOx") in inv._gdf_columns
            or (cat in inv.gdfs and "NOx" in inv.gdfs[cat].columns)
        }
    else:
        raise TypeError(f"NOX_TO_NO2 must be a float or dict, not {type(NOX_TO_NO2)}.")

    return speciate_inventory(inv, speciation_dict, drop=drop)


def merge_substances(
    inv: Inventory,
    substances: dict[str, list[str]],
    drop: bool = True,
    inplace: bool = False,
) -> Inventory:
    """Merge substances in an inventory.

    :arg inv: The inventory to merge substances.
    :arg substances: A dict with the substances to merge.
        The keys are the new substance names.
        The values are the list of substances to merge.
    :arg drop: Whether to drop the merged substances.
    :arg inplace: Whether to modify the inventory in place.

    :returns: The inventory with the merged substances.
    """
    if inplace:
        new_inv = inv
    else:
        new_inv = inv.copy()
    inv_subs = inv.substances

    # Check that the substances merge is valid
    # New substances should not be contained in merges of other substances
    for new_substance, old_substances in substances.items():
        for old_substance in old_substances:
            if old_substance not in inv_subs:
                raise KeyError(f"{old_substance} is not in {inv} .")
            if old_substance in substances.keys() and new_substance != old_substance:
                raise ValueError(
                    f"Cannot merge {old_substance} into {new_substance} because"
                    f" {old_substance} is merged."
                )

    for new_substance, old_substances in substances.items():
        for cat in inv.categories:
            cols_to_merge = []
            for old_substance in old_substances:
                if (cat, old_substance) in inv._gdf_columns:
                    cols_to_merge.append((cat, old_substance))
            if not cols_to_merge:
                continue
            # Merge the gdf
            total = new_inv.gdf.loc[:, cols_to_merge].sum(axis=1)
            new_inv.gdf[(cat, new_substance)] = total
            if drop:
                new_inv.gdf.drop(
                    columns=[c for c in cols_to_merge if c[1] != new_substance],
                    inplace=True,
                )

    # Apply the same operation to the gdfs
    for cat, gdf in inv.gdfs.items():
        cols_to_merge = [old_s for old_s in old_substances if old_s in gdf.columns]
        if not cols_to_merge:
            continue
        new_inv.gdfs[cat][new_substance] = gdf.loc[:, cols_to_merge].sum(axis=1)
        if drop:
            new_inv.gdfs[cat].drop(
                columns=[s for s in cols_to_merge if s != new_substance], inplace=True
            )

    new_inv.history.append(f"Merged substances with {substances}.")

    return new_inv



"""Example of remapping the TNO inventory to the ICON grid and exporting it to OEM."""
# %%
from pathlib import Path

from emiproc.grids import WGS84, ICONGrid, WGS84_PROJECTED
from emiproc.inventories.tno import TNO_Inventory
from emiproc.inventories.categories_groups import TNO_2_GNFR
from emiproc.inventories.utils import group_categories
from emiproc.regrid import remap_inventory
from emiproc.exports.icon import export_icon_oem, TemporalProfilesTypes


# %% Declare path of the TNO netcdf file
nc_file = r"C:\Users\coli\Documents\emiproc\files\TNO_6x6_GHGco_v4_0\TNO_GHGco_v4_0_year2018.nc"

# %% Load the inventory to an object
inv = TNO_Inventory(nc_file)


# %% Load the icon grid
grid_file = Path(r"C:\Users\coli\Documents\ZH-CH-emission\icon_europe_DOM01.nc")
icon_grid = ICONGrid(grid_file)

# %%
# Convert to a planar crs, required for surface conserving remapping
inv.to_crs(WGS84_PROJECTED)

# %% Remap the inventory to the icon grid
remaped_tno = remap_inventory(
    inv, icon_grid, grid_file.parent / f".emiproc_remap_tno2{grid_file.stem}"
)

# %% Group the categories to the GNFR (Mainly renaming them)
groupped = group_categories(
    inv=remaped_tno,
    categories_group=TNO_2_GNFR
)

# %% Export the invenotry to OEM

output_dir = grid_file.with_name(f"{grid_file.stem}_with_tno_emissions")

export_icon_oem(
    inv=groupped,
    icon_grid_file=grid_file,
    output_dir=output_dir,
    temporal_profiles_type=TemporalProfilesTypes.THREE_CYCLES,
    # Following parameters are for HOUR_OF_YEAR profiles
    #temporal_profiles_type=TemporalProfilesTypes.HOUR_OF_YEAR,
    #year=2020,
)

print(f"Exported to {output_dir}")

# %%



"""plots netcdf raster files"""

# %% Plot the total emissions
from pathlib import Path
import matplotlib as mpl
from emiproc.plots import nclcmaps
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt

# %%

cmap = nclcmaps.cmap("WhViBlGrYeOrRe")
mpl.style.use("default")
from matplotlib.colors import LogNorm, SymLogNorm


# %%
file = Path(
    r"C:\Users\coli\Documents\ZH-CH-emission\output_files\mapluft_rasters\zurich_cropped_Fsplit_100x100_mapLuft_2022_v2024_v1.7.nc"
)
ds = xr.load_dataset(file)
x_min, x_max, y_min, y_max = min(ds["x"]), max(ds["x"]), min(ds["y"]), max(ds["y"])
ds
images_dir = file.with_suffix(".plots")
images_dir.mkdir(exist_ok=True)

substances = [
    "CO2_bio",
    "CO2_fos",
    "CO",
    "BC",
    # "PM10ex",
    # "PM10non",
    # "PM25ex",
    # "PM25non",
    "SO2",
    "NOx",
    "N2O",
    #"NH3",
    "CH4",
    # "VOC",
    # "benzene",
]

categories = list(ds["source_category"].values)

# %% Make plots with the total emissions

fig, ax = plt.subplots()

colors_of_category = {
    "GNFR_A": "olive",
    "GNFR_B": "brown",
    "GNFR_C": "tomato",
    "GNFR_D": "cyan",
    "GNFR_E": "blue",
    "GNFR_F-cars": "orange",
    "GNFR_F-heavy_duty": "darkorange",
    "GNFR_F-light_duty": "peru",
    "GNFR_F-two_wheels": "sandybrown",
    "GNFR_G": "hotpink",
    "GNFR_I": "purple",
    "GNFR_J": "gray",
    "GNFR_K": "teal",
    "GNFR_L": "maroon",
    "GNFR_O": "green",
    "GNFR_R": "lightgreen",
}
co2_substances = ["CO2_bio", "CO2_fos"]
for i, substance in enumerate(substances):

    total_emissions_per_sector = ds[f"emi_{substance}_total"]
    if substance in co2_substances:
        # Sum over all the CO2
        total_sum = np.sum([ds[f"emi_{s}_all_sectors"].sum().values for s in co2_substances])
    else:
        total_sum = total_emissions_per_sector.sum().values
    # Stack the emissions from each category
    cumsum = 0

    for cat in sorted(categories):
        da = total_emissions_per_sector.sel(source_category=cat)
        emissions = da.values
        emissions = np.nan_to_num(emissions, nan=0.0)
        # Remove the nan values
        kwargs = {}
        if i == 0:
            kwargs["label"] = cat
        ax.bar(
            x=i,
            bottom=cumsum / total_sum,
            height=emissions / total_sum,
            width=0.8,
            color=colors_of_category[cat],
            **kwargs,
        )
        cumsum += emissions

ax.set_xticks(range(len(substances)))
ax.set_xticklabels(substances, rotation=45)
ax.legend(loc='lower center', bbox_to_anchor=(0.5, 1.00), ncol=3)
ax.set_ylabel("Fraction of total emissions")

fig.tight_layout()
file_name = images_dir / f"total_emissions_fractions_{file.stem}"
fig.savefig(file_name.with_suffix(".png"))


# %% plots all substances


q_max = 0.001
q_min = 0.2
for substance in substances:

    for cat in categories + ["__total__"]:
        if cat == "__total__":
            da = ds[f"emi_{substance}_all_sectors"]
            cat_sub_emissions = da.to_numpy()
        else:
            da = ds[f"emi_{substance}"]
            cat_sub_emissions = da.sel(source_category=cat).to_numpy()

        # cax = fig.add_axes([0.87, 0.15, 0.02, 0.7])

        emission_non_zero_values = cat_sub_emissions[cat_sub_emissions > 0]
        if not np.any(emission_non_zero_values):
            # No emissions at all
            continue

        fig, axes = plt.subplots(
            figsize=(10, 8),
            ncols=2,
            # Set the colorbar to a smaller size
            gridspec_kw={"width_ratios": [1, 0.05]},
        )
        ax, cax = axes
        norm = LogNorm(
            # vmin=1e-1,
            vmin=np.quantile(emission_non_zero_values, q_min),
            # vmax=1e4,
            vmax=np.quantile(emission_non_zero_values, 1 - q_max),
        )

        ax.clear()
        im = ax.imshow(
            cat_sub_emissions,
            norm=norm,
            cmap=cmap,
            # cmap="inferno",
            # extent=[wgs_84_extents[1],wgs_84_extents[3], wgs_84_extents[0],wgs_84_extents[2] ]
            extent=[x_min, x_max, y_min, y_max],
        )
        ax.xaxis.set_major_formatter("{x:6.0f}")
        ax.yaxis.set_major_formatter("{x:6.0f}")
        ax.set_title(
            f"Total {substance} {cat} from Zurich: "
            f"{cat_sub_emissions.sum():.2} "
            f"{da.units}"
        )
        cax.clear()

        fig.colorbar(
            im,
            extend="both",
            extendfrac=0.02,
            label=da.units,
            cax=cax,
        )
        fig.tight_layout()
        file_name = images_dir / f"raster_{substance}_{cat}_{file.stem}"
        fig.savefig(file_name.with_suffix(".png"))
        # fig.savefig(file_name.with_suffix(".pdf"))

        # close
        plt.close(fig)

# %%



from os import PathLike

import geopandas as gpd
import pandas as pd
import xarray as xr

from emiproc.grids import BoundingBox, RegularGrid
from emiproc.inventories import Inventory
from emiproc.profiles.temporal_profiles import (
    CompositeTemporalProfiles,
    DayOfYearProfile,
    get_leap_year_or_normal,
)
from emiproc.profiles.utils import ratios_dataarray_to_profiles
from emiproc.utilities import SEC_PER_YR


class GFAS_Inventory(Inventory):
    def __init__(self, *args, **kwargs):
        raise DeprecationWarning(
            "GFAS_Inventory has been updated and renamed to GFAS. Please use GFAS instead."
        )


class GFAS(Inventory):
    """The GFAS inventory.

    Contains gridded data of forest fires from the Copernicus Atmosphere Monitoring
    Service (CAMS).

    You can access the data at
    `CAMS global biomass burning emissions based on fire radiative power
    <https://ads.atmosphere.copernicus.eu/cdsapp#!/dataset/cams-global-fire-emissions-gfas>`_

    """

    grid: RegularGrid

    def __init__(
        self,
        nc_file: PathLike,
        variables: list[str] = [],
        bbox: BoundingBox | None = None,
    ):
        """Create a GFAS inventory.

        The GFAS nc file contains gridded data of forest fires.
        By default, all variables in the nc file are included. If you want to include
        only a subset of the variables, you can specify them in the variables argument.

        :param nc_file: The path to the netCDF file.
            Make sure the file contains one year of data
        :param variables: A list of variables to include in the inventory.
        :param bbox: A bounding box to subset the data.
            `[minx, miny, maxx, maxy]`

        """
        super().__init__()
        ds = xr.open_dataset(nc_file)

        self.year = pd.Timestamp(ds["valid_time"].values[0]).year

        # Check that the file contains one year of data
        expected_profile = get_leap_year_or_normal(DayOfYearProfile, year=self.year)
        if expected_profile.size != ds["valid_time"].size:
            raise ValueError(
                f"Expected {expected_profile.size} timesteps for year {self.year}, "
                f"but got {ds['valid_time'].size} timesteps in the file."
                " Make sure the file contains one (and only one) full year of data."
            )

        # Apply optional bounding box to subset the data
        if bbox:
            lon_mask = (ds["longitude"] >= bbox[0]) & (ds["longitude"] <= bbox[2])
            lat_mask = (ds["latitude"] >= bbox[1]) & (ds["latitude"] <= bbox[3])
            ds = ds.sel(longitude=lon_mask, latitude=lat_mask)

        if not variables:
            variables = list(ds.data_vars.keys())
        var_2_emiproc = {var: var.replace("fire", "").upper() for var in variables}

        profiles = {
            sub: ds[key].expand_dims(substance=[sub])
            for key, sub in var_2_emiproc.items()
        }

        self.grid = RegularGrid.from_centers(
            x_centers=ds["longitude"].values,
            y_centers=ds["latitude"].values,
            name="GFAS",
            rounding=2,
            # Custom projection, to correct for the lon range of the data
            crs="+proj=longlat +datum=WGS84 +no_defs +type=crs +lon_wrap=180",
        )

        self.gdfs = {}

        # Reshape the data to the regular grid
        da_profiles: xr.DataArray = (
            xr.concat(list(profiles.values()), dim="substance")
            .stack(cell=("longitude", "latitude"))
            .drop_vars(["longitude", "latitude"])
            .rename({"valid_time": "ratio"})
        )

        def process_substance(sub):
            return (
                da_profiles.sel(substance=sub).mean("ratio")
                # Convert from kg m-2 s-1 to kg/yr
                * SEC_PER_YR
                * self.cell_areas.reshape(-1)
            )

        self.gdf = gpd.GeoDataFrame(
            {("gfas", sub): process_substance(sub) for sub in var_2_emiproc.values()},
            geometry=self.grid.gdf.geometry,
        )

        ratios, indices = ratios_dataarray_to_profiles(da_profiles)

        profiles = CompositeTemporalProfiles.from_ratios(
            ratios,
            [get_leap_year_or_normal(DayOfYearProfile, year=self.year)],
            rescale=True,
        )

        self.set_profiles(profiles, indices)



"""Emiproc provides tools to help generate human respiration emissions.



"""
from __future__ import annotations

from os import PathLike
import geopandas as gpd
from emiproc.grids import RegularGrid
from emiproc.inventories import Inventory, Category, Substance, CatSub
from emiproc.utilities import DAY_PER_YR


class EmissionFactor:
    """Emissions factors are how much of a compound is produced per individual per day.

    Units are kg/day (kg of the compounds)
    """

    # Very simple estimation for aproximate results
    ROUGH_ESTIMATON = 1.0
    CO2_ROUGH_ESTIMATON = 1.0

    # 2.1 g / year / person 
    # https://doi.org/10.1016/S0048-9697(97)00267-2
    N2O_MITSUI_ET_ALL = 2.1e-3 / DAY_PER_YR

    # Breath: 2.3 mmol / day / person 
    # Flatus: 7.0 mmol / day / person 
    # https://doi.org/10.1016/j.atmosenv.2019.116823
    # Multiply by the molar mass of CH4 and correct units
    CH4_POLAG_KEPPLER = (2.3 + 7.0) * 1e-3 * 16e-3 


def load_data_from_quartieranalyse(
    file: PathLike,
    grid: RegularGrid | None = None,
) -> gpd.GeoDataFrame:
    """Load the data required from the quartieranalyse file from zurich.

    :arg file: The Path to the Quartier data file.
    :arg grid: Optionally the grid of the simulation. This helps to
        load the data.

    :return: A geodataframe containing columns:
        geometry: The shapes of the different zones
        people_working: The number of people who work in these zones
        people_living: The number of people who live in these zones

    """
    kwargs = {}
    if grid:
        kwargs["bbox"] = grid.bounds

    gdf = gpd.read_file(file, **kwargs)
    gdf_interest = gdf[["U_EINW", "U_BESCH", "geometry"]]

    # Replace missing values
    gdf_interest = gdf_interest.replace(-999.0, 0)

    gdf_interest = gdf_interest.rename(
        columns={
            "U_BESCH": "people_working",
            "U_EINW": "people_living",
        }
    )

    return gdf_interest


def people_to_emissions(
    people_gdf: gpd.GeoDataFrame,
    time_ratios: dict[Category, float],
    emission_factor:  dict[CatSub, float] | float = EmissionFactor.ROUGH_ESTIMATON,
    output_gdfs: bool = False,
    name: str = "human_respiration",
    substance: Substance = "CO2",
) -> Inventory:
    """Get human respiration emissions.

    Convert the number of people living in different areas
    to annual emission of CO2 from human respiration.

    Different categories can be provided in the input. (ex. working, living, etc.)

    The formula used is quite simple:

    .. math::
        \\text{emissions} [kg/y/shape] =
        \\text{emission factor} [kg/p/d]
        * \\text{people} [p/shape]
        * \\text{time ratio} [-]
        * \\text{days per year} [d/y]

    :arg people_gdf: A geodataframe containing the number of people for
        each geometry/shape/row. Each column must be named after the category of people.
    :arg time_ratios: The ratio of time that people spend for each category.
        The sum of all the ratios must be 1.
        Ex: `{'working': 0.3, 'living': 0.7}`.
    :arg emission_factor: The emission factor to use for each of the activities.
        Can be a single value or a dict with the same keys as the categories.
    :arg output_gdfs: Whether the output inventory should contain the emission
        data in gdfs instead of in the gdf(default).
    :arg name: The name of the inventory.
    :arg substance: The substance name of the emissions. 

    :return: The Inventory containing human emissions.
    """

    categories = list(time_ratios.keys())
    if isinstance(emission_factor, (int, float)):
        # Apply the emission to all categories
        emission_factor = {(cat, substance): emission_factor for cat in categories}
    else:
        # Check that the keys are correct
        for key in emission_factor.keys():
            if not isinstance(key, tuple) or len(key) != 2:
                raise ValueError(f"Invalid key {key = } in emission_factor. Must be a tuple of (category, substance).")
            cat, sub = key
            if cat not in categories:
                raise ValueError(f"Invalid category {cat = } in emission_factor. Must be one of {categories = }")
    if not sum(time_ratios.values()) == 1:
        raise ValueError(
            f"The time ratios must sum up to 1, got {time_ratios.values() = }"
        )

    yearly_emissions = {
        # Calculate the yearly emissions for each of the categories
        # (kg/p/day) * (p/shape) * (-) * (day/y) = (kg/y/shape)
        catsub: factor * people_gdf[catsub[0]] * time_ratios[catsub[0]] * DAY_PER_YR
        for catsub, factor in emission_factor.items()
    }

    inv_kwargs = {}
    if output_gdfs:
        inv_kwargs["gdfs"] = {
            catsub[0]: gpd.GeoDataFrame(
                {catsub[1]: yearly_emissions[catsub]}, geometry=people_gdf.geometry
            )
            for catsub in emission_factor.keys()
        }
    else:
        # Use the main gdf
        inv_kwargs["gdf"] = gpd.GeoDataFrame(
            {catsub: yearly_emissions[catsub] for catsub in emission_factor.keys()},
            geometry=people_gdf.geometry,
        )

    return Inventory.from_gdf(**inv_kwargs, name=name)



import pytest 
from emiproc.inventories import  EmissionInfo, Inventory
from emiproc.regrid import remap_inventory
from emiproc.tests_utils.test_inventories import inv_with_pnt_sources, inv_with_gdfs_bad_indexes, inv
from emiproc.tests_utils.test_grids import regular_grid, gpd_grid
from emiproc.tests_utils.temporal_profiles import three_composite_profiles, indexes_inv_catsubcell

from emiproc.inventories.utils import get_total_emissions
from emiproc.utilities import total_emissions_almost_equal




def test_remap():

    grid = regular_grid
    grid.crs = None
    remaped_inv = remap_inventory(inv_with_pnt_sources, grid)

    # Check the grid size
    assert remaped_inv.gdf.shape[0] == len(grid.gdf)
    assert len(remaped_inv.gdfs) == 0, "No point sources should be left"

    # Check the total emissions dictionaries are the same (grid is larger than inventory)
    total_inv = get_total_emissions(inv_with_pnt_sources)
    total_remapped = get_total_emissions(remaped_inv)
    assert total_emissions_almost_equal(total_inv, total_remapped)


def test_remap_keep_shapes():


    grid = regular_grid
    grid.crs = None
    remaped_inv = remap_inventory(inv_with_pnt_sources, grid, keep_gdfs=True)

    # Check the grid size
    assert remaped_inv.gdf.shape[0] == len(grid.gdf)
    # Check the point sources
    # The dataframes should be the same 
    assert list(remaped_inv.gdfs.keys()) == list(inv_with_pnt_sources.gdfs.keys())
    for key in remaped_inv.gdfs.keys():
        assert remaped_inv.gdfs[key].shape == inv_with_pnt_sources.gdfs[key].shape
        assert list(remaped_inv.gdfs[key].columns) == list(inv_with_pnt_sources.gdfs[key].columns)
        assert remaped_inv.gdfs[key].geometry.equals(inv_with_pnt_sources.gdfs[key].geometry)
        # Check all values 
        for col in remaped_inv.gdfs[key].columns:
            assert remaped_inv.gdfs[key][col].equals(inv_with_pnt_sources.gdfs[key][col])

    # Check the total emissions dictionaries are the same (grid is larger than inventory)
    total_inv = get_total_emissions(inv_with_pnt_sources)
    total_remapped = get_total_emissions(remaped_inv)
    assert total_emissions_almost_equal(total_inv, total_remapped)




def test_remap_different_grids():
    for grid in [regular_grid, gpd_grid]:
        try:
            grid.crs = None
            remaped_inv = remap_inventory(inv_with_pnt_sources, grid=grid)
        except Exception as e:
            raise AssertionError(f"Remapping failed for grid {grid.name}") from e



def test_remap_with_gdf_wrong_indices():
    """Test that the remap_inventory function works also if the indices in the gdfs are bad."""

    inv = inv_with_gdfs_bad_indexes

    regular_grid.crs = None
    remapped = remap_inventory(inv, regular_grid)


def test_remap_inv_with_profiles():

    this_inv = inv.copy()

    this_inv.set_profiles(three_composite_profiles, indexes_inv_catsubcell)

    remapped_inv = remap_inventory(this_inv, gpd_grid)

    assert len(remapped_inv.t_profiles_groups) > 1, "There should be more than one profiles"
    assert remapped_inv.t_profiles_indexes.dims == indexes_inv_catsubcell.dims


import logging
from pathlib import Path
from urllib.request import urlretrieve

import xarray as xr

from emiproc import FILES_DIR
from emiproc.grids import ICONGrid

ICON_GRID_DIR = FILES_DIR / "test" / "icon_grid"

ICON_GRID_DIR.mkdir(parents=True, exist_ok=True)

logger = logging.getLogger(__name__)


ICON_GRID_REPO = "http://icon-downloads.mpimet.mpg.de/grids/public/edzw/"

SIMPLE_ICON_GRID_PATH = ICON_GRID_DIR / "icon_grid_0055_R02B05_N.nc"


def download_test_grid(grid_filename: str):
    """Download a test grid from the ICON grid repository."""

    url = ICON_GRID_REPO + grid_filename
    logger.info("Downloading test grid from %s", url)
    urlretrieve(url, ICON_GRID_DIR / Path(url).name)


def get_test_grid(
    grid_path: Path = SIMPLE_ICON_GRID_PATH,
) -> ICONGrid:
    """Return the path of the test grid."""

    if not grid_path.exists():
        download_test_grid(grid_path.name)
    return ICONGrid(grid_path)


if __name__ == "__main__":
    print(get_test_grid())



from pathlib import Path


import xarray as xr
import numpy as np
import geopandas as gpd
from shapely.creation import polygons

from emiproc.grids import WGS84_PROJECTED, GeoPandasGrid
from emiproc.inventories import Inventory
from emiproc.profiles.temporal_profiles import (
    CompositeTemporalProfiles,
    DayOfYearProfile,
    MounthsProfile,
)


class SaunoisInventory(Inventory):
    """Inventory based on Saunois estimates of methane emissions.

    https://doi.org/10.5194/essd-12-1561-2020

    You can download the data there but note that the current implementation
    had some other files as sources, but very similar.
    https://www.icos-cp.eu/GCP-CH4-2024

    """

    def __init__(self, saunois_files: list[Path]):
        """Initialize the inventory.


        Parameters
        ----------

        saunois_files :
            List of paths to the Saunois files.
            Here each netcdf is named after the category.
            If you donwload from the ICOS website, you will have to rename the files.
            Or change the code for this inventory.
        """
        super().__init__()

        da = xr.concat(
            [
                xr.open_dataset(file)["flux"]
                .rename(file.stem)
                .expand_dims(category=[file.stem])
                for file in saunois_files
            ],
            dim="category",
        )

        # Drop the lev dimension and add the substance dimension
        assert da["lev"].size == 1
        da = da.squeeze("lev").expand_dims(substance=["CH4"])

        # Set the coords to be str
        da["substance"] = da["substance"].astype(str)
        da["category"] = da["category"].astype(str)

        # replace the lat lon by cell
        da_stacked_all = da.stack(cell=("lon", "lat"))
        da_stacked = da_stacked_all.drop_vars(["lat", "lon"])
        # Use a simple integer index for the cell
        da_stacked["cell"] = np.arange(da_stacked.sizes["cell"])

        # Make the geometry
        lon = da_stacked_all["lon"].values
        lat = da_stacked_all["lat"].values

        lon_range = da["lon"].values
        lat_range = da["lat"].values

        d_lon = np.round(np.diff(lon_range)[0], 8)
        d_lat = np.round(np.diff(lat_range)[0], 8)
        # Ensure the spacing is constant
        assert np.allclose(
            np.diff(lon_range), d_lon
        ), "Longitude spacing is not constant"
        assert np.allclose(
            np.diff(lat_range), d_lat
        ), "Latitude spacing is not constant"

        # Reconstruct the grid vertices
        coords = np.array(
            [
                # Bottom left
                [
                    lon - d_lon / 2,
                    lat - d_lat / 2,
                ],
                # Bottom right
                [
                    lon + d_lon / 2,
                    lat - d_lat / 2,
                ],
                # Top right
                [
                    lon + d_lon / 2,
                    lat + d_lat / 2,
                ],
                # Top left
                [
                    lon - d_lon / 2,
                    lat + d_lat / 2,
                ],
            ]
        )

        coords = np.rollaxis(coords, -1, 0)

        geometry = gpd.GeoSeries(polygons(coords), crs="WGS84")
        self.grid = GeoPandasGrid(geometry, shape=(len(lon_range), len(lat_range)))

        # Unit conversion
        # Units are gCH4/m2/day
        # "g CH4 m-2 d-1" -> "kg / year / cell"
        # To convert from day to year, we have to multiplly each month by the number of days in the month and then sum the months totals
        da_stacked_total = (
            da_stacked
            * xr.DataArray(
                np.array([31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]),
                dims="time",
                coords={"time": da_stacked.coords["time"]},
            )
        ).sum(dim="time")
        # kg/g * m2/cell
        converstion_factor = 1e-3 * geometry.to_crs(WGS84_PROJECTED).area
        da_total = da_stacked_total * converstion_factor.values

        # Convert to pandas
        df = (
            da_total.stack(catsub=("category", "substance"))
            .drop_vars(["cell"])
            .to_pandas()
        )
        self.gdf = gpd.GeoDataFrame(df, geometry=geometry)
        self.gdfs = {}

        # Generate the profiles
        da_profiles = da_stacked.stack(profiles=("substance", "category", "cell"))
        # Convert to ratios
        da_ratios = (da_profiles / da_profiles.sum(dim="time")).fillna(0.0)
        mask_valid = da_ratios.sum(dim="time") != 0

        # Create the profiles indexes
        da_profiles_indexes = da_ratios.sum(dim="time")
        da_profiles_indexes.values = -np.ones(da_profiles_indexes.shape, dtype=int)

        # Set the values to linear indices
        da_valid_profiles = da_ratios.sel(profiles=mask_valid)

        # Many profiles are exactly the same, so we will simplify the profiles by grouping the same profiles
        unique_profiles, unique_indices = np.unique(
            da_valid_profiles.values, axis=-1, return_inverse=True
        )

        da_profiles_indexes.loc[mask_valid] = unique_indices
        profiles_indexes = da_profiles_indexes.unstack()

        self.set_profiles(
            profiles=CompositeTemporalProfiles.from_ratios(
                unique_profiles.T, types=[MounthsProfile]
            ),
            indexes=profiles_indexes.drop_vars("lev"),
        )



from __future__ import annotations

import pytest
from emiproc import FILES_DIR
from datetime import datetime
import numpy as np
import xarray as xr
from emiproc.tests_utils import temporal_profiles
from emiproc.profiles.temporal_profiles import (
    DailyProfile,
    SpecificDay,
    SpecificDayProfile,
    days_of_specific_day,
    ensure_specific_days_consistency,
    from_yaml,
    create_scaling_factors_time_serie,
    profile_to_scaling_factors,
    get_profile_da,
)


def test_ensure_specific_days_consistency_non_specific():
    out = ensure_specific_days_consistency(temporal_profiles.oem_const_profile)
    assert len(out) == len(temporal_profiles.oem_const_profile)

    out = ensure_specific_days_consistency(
        [SpecificDayProfile(specific_day=SpecificDay("monday"))]
    )
    assert len(out) == 7  # Contains the full week

    half_double_ratios = np.concatenate((np.full(12, 0.5), np.full(12, 1.5))) / 24
    thirds_triple_ratios = np.concatenate((np.full(12, 1 / 3), np.full(12, 5 / 3))) / 24
    quaters_ratios = np.concatenate((np.full(12, 1 / 4), np.full(12, 7 / 4))) / 24
    out = ensure_specific_days_consistency(
        [
            SpecificDayProfile(
                ratios=half_double_ratios, specific_day=SpecificDay.MONDAY
            ),
            SpecificDayProfile(
                ratios=thirds_triple_ratios, specific_day=SpecificDay.WEEKEND
            ),
            SpecificDayProfile(ratios=quaters_ratios, specific_day=SpecificDay.SUNDAY),
            DailyProfile(),
        ],
    )
    for p in out:
        assert isinstance(p, SpecificDayProfile)
        assert len(days_of_specific_day(p.specific_day)) == 1
        if p.specific_day == SpecificDay.MONDAY:
            equalt_to = half_double_ratios
        elif p.specific_day == SpecificDay.SATURDAY:
            equalt_to = thirds_triple_ratios
        elif p.specific_day == SpecificDay.SUNDAY:
            equalt_to = quaters_ratios
        else:
            equalt_to = np.full(24, 1.0) / 24
        np.testing.assert_array_equal(
            p.ratios.reshape(-1), equalt_to, err_msg=f"Failed for {p.specific_day=}"
        )


@pytest.mark.parametrize(
    "profiles",
    [
        temporal_profiles.oem_const_profile,
        from_yaml(FILES_DIR / "profiles" / "yamls" / "heat.yaml"),
        from_yaml(FILES_DIR / "profiles" / "yamls" / "human_home.yaml"),
        from_yaml(FILES_DIR / "profiles" / "yamls" / "human.yaml"),
        from_yaml(FILES_DIR / "profiles" / "yamls" / "ship.yaml"),
        from_yaml(FILES_DIR / "profiles" / "yamls" / "heavy.yaml"),
        from_yaml(FILES_DIR / "profiles" / "yamls" / "light.yaml"),
    ],
)
def test_create_scaling_factors_time_serie(profiles):
    start = datetime(2018, 1, 1)
    end = datetime(2019, 1, 1)
    ts = create_scaling_factors_time_serie(start, end, profiles)

    assert np.isclose(ts.mean(), 1.0, atol=0.1)


@pytest.mark.parametrize(
    "profile",
    [
        temporal_profiles.daily_test_profile,
        temporal_profiles.weekly_test_profile,
        temporal_profiles.mounths_test_profile,
    ],
)
def test_profile_da(profile):

    da = get_profile_da(profile, year=2018)
    assert isinstance(da, xr.DataArray)



import pytest
import xarray as xr
import numpy as np

from emiproc.profiles.operators import (
    combine_profiles,
    country_to_cells,
    get_weights_of_gdf_profiles,
    group_profiles_indexes,
    weighted_combination,
)
from emiproc.profiles.temporal_profiles import (
    DailyProfile,
    WeeklyProfile,
    CompositeTemporalProfiles,
)
from emiproc.tests_utils import temporal_profiles, vertical_profiles
from emiproc.tests_utils.temporal_profiles import (
    TEST_COPENICUS_PROFILES,
    get_random_profiles,
    indexes_african_2d,
    indexes_african_simple,
    read_test_copernicus,
)
from emiproc.tests_utils.test_grids import regular_grid_africa
from emiproc.tests_utils.vertical_profiles import (
    get_random_profiles as get_random_profiles_vertical,
)


def test_reading_copernicus():
    profiles = read_test_copernicus()

    assert len(profiles) == len(TEST_COPENICUS_PROFILES)


def test_weighted_combination():
    weighted_combination([DailyProfile(), DailyProfile()], [0.5, 0.5])


def test_weighted_combination_fails_on_different_profiles_types():
    pytest.raises(
        TypeError, weighted_combination, [DailyProfile(), WeeklyProfile()], [0.5, 0.5]
    )


def test_weighted_combination_fails_on_wrong_weights_lenghts():
    pytest.raises(
        ValueError, weighted_combination, [DailyProfile(), DailyProfile()], [0.5]
    )


def test_combine_profiles_single_dimension():
    new_profiles, new_indices = combine_profiles(
        profiles=vertical_profiles.get_random_profiles(2),
        profiles_indexes=vertical_profiles.single_dim_profile_indexes,
        weights=vertical_profiles.single_dim_weights,
        dimension="category",
    )


def test_combine_profiles():
    weights = get_weights_of_gdf_profiles(
        vertical_profiles.inv.gdf, vertical_profiles.corresponding_2d_profiles
    )
    new_profiles, new_indices = combine_profiles(
        profiles=temporal_profiles.three_profiles,
        profiles_indexes=vertical_profiles.corresponding_2d_profiles,
        weights=weights,
        dimension="category",
    )
    assert "category" not in new_indices.dims


def test_group_profiles_with_time_profiles():
    weights = get_weights_of_gdf_profiles(
        vertical_profiles.inv.gdf, vertical_profiles.corresponding_2d_profiles
    )
    new_profiles, new_indices = group_profiles_indexes(
        profiles=temporal_profiles.three_profiles,
        profiles_indexes=vertical_profiles.corresponding_2d_profiles,
        indexes_weights=weights,
        categories_group=vertical_profiles.inv_groups_dict,
        groupping_dimension="category",
    )
    assert "category" in new_indices.dims


def test_get_random_profiles():
    p = get_random_profiles(3)
    assert len(p) == 3


def test_get_random_profies_vertical():
    p = get_random_profiles_vertical(5)
    assert len(p) == 5


@pytest.mark.parametrize(
    "profiles, indexes",
    [
        (
            get_random_profiles_vertical(indexes_african_simple.max().values + 1),
            indexes_african_simple,
        ),
        (
            get_random_profiles(indexes_african_simple.max().values + 1),
            indexes_african_simple,
        ),
        (
            get_random_profiles_vertical(indexes_african_2d.max().values + 1),
            indexes_african_2d,
        ),
        (
            get_random_profiles(indexes_african_2d.max().values + 1),
            indexes_african_2d,
        ),
    ],
)
def test_countries_to_cells(profiles, indexes: xr.DataArray):
    grid = regular_grid_africa

    if isinstance(profiles, list):
        profiles = CompositeTemporalProfiles(profiles)

    new_profiles, new_indexes = country_to_cells(profiles, indexes, grid)

    assert "cell" in new_indexes.dims
    assert "country" not in new_indexes.dims

    if isinstance(profiles, CompositeTemporalProfiles):
        assert profiles.types == new_profiles.types
    # test for some cells that we now what it should be
    # Full in MRT, profiles should be the same as the original
    np.testing.assert_almost_equal(
        new_profiles.scaling_factors[new_indexes.sel(cell=78).drop_vars("cell")],
        profiles.scaling_factors[indexes.sel(country="MRT").drop_vars("country")],
    )
    # THis is shared between SEN and ocean, so only in SEN for the profile
    np.testing.assert_almost_equal(
        new_profiles.scaling_factors[new_indexes.sel(cell=26).drop_vars("cell")],
        profiles.scaling_factors[indexes.sel(country="SEN").drop_vars("country")],
    )

    # This is just ocean
    assert np.all(new_indexes.sel(cell=0).drop_vars("cell") == -1)


if __name__ == "__main__":
    pytest.main([__file__])
    # test_group_profiles_with_time_profiles()



import pytest
import numpy as np
import json
import emiproc
from emiproc.profiles.temporal_profiles import (
    TemporalProfile,
    DailyProfile,
    WeeklyProfile,
    MounthsProfile,
    from_csv,
    from_yaml,
    read_temporal_profiles,
    to_yaml,
)

from emiproc.profiles.vertical_profiles import read_vertical_profiles


def test_saving_and_loading_yamls_temporal_profiles():
    yaml_dir = emiproc.FILES_DIR / "profiles" / "yamls"
    yaml_profiles = {}
    for yml_file in yaml_dir.glob("*.yaml"):
        yaml_profiles[yml_file.stem] = from_yaml(yml_file)
    # We can also save time profiles to yaml
    loaded = {}
    for categorie in yaml_profiles.keys():
        yaml_file = emiproc.FILES_DIR / "outputs" / f"test_{categorie}.yaml"
        to_yaml(yaml_profiles[categorie], yaml_file)
        # Make sure we can load these yamls
        loaded[categorie] = from_yaml(yaml_file)
        # TODO: check that the loaded yaml is the same as the original one


def test_load_csv_profiles():
    copernicus_profiles_dir = emiproc.FILES_DIR / "profiles" / "copernicus"

    profiles = ["hour_in_day", "day_in_week", "month_in_year"]
    profiles = {
        p: from_csv(copernicus_profiles_dir / f"timeprofiles-{p}.csv") for p in profiles
    }


def test_read_simple():
    """Test a simple file.

    Category,Substance,10,20,50
    blek,CO2,0.1,0.2,0.7
    liku,CO2,0.1,0.3,0.6
    liku,CH4,0,0.2,0.8
    blek,N20,0.2,0.4,0.4
    """
    proffile = emiproc.FILES_DIR / "test/profiles/simple_vertical/test_file.csv"

    profiles, indexes = read_vertical_profiles(proffile)

    # Test now the indexes are correct
    expect_present_dims = ["category", "substance"]
    for dim in expect_present_dims:
        assert dim in indexes.dims
    assert len(indexes.dims) == len(expect_present_dims)

    assert len(indexes.coords["category"]) == 2
    assert len(indexes.coords["substance"]) == 3
    # Test the indexes are correct
    # Note: this might be wrong if one day we change the order of indexing the profiles
    # Now we assume same order as in file
    assert indexes.loc["blek", "CO2"] == 0
    assert indexes.loc["liku", "CO2"] == 1
    assert indexes.loc["liku", "CH4"] == 2
    assert indexes.loc["blek", "N20"] == 3
    # Missing should be -1
    assert indexes.loc["blek", "CH4"] == -1
    assert indexes.loc["liku", "N20"] == -1

    # Test the profiles are correct
    assert len(profiles) == 4
    assert all(profiles.height == np.array([10, 20, 50]))


# Template to do the abovee test


@pytest.mark.parametrize(
    "name,profiles_dir,n_profiles,expected_dict",
    [
        (
            "Simple file",
            emiproc.FILES_DIR / "test/profiles/simple_vertical/test_file.csv",
            4,
            {
                '{"category":"blek","substance":"CO2"}': [0.1, 0.2, 0.7],
                '{"category":"liku","substance":"CO2"}': [0.1, 0.3, 0.6],
                '{"category":"liku","substance":"CH4"}': [0, 0.2, 0.8],
                '{"category":"blek","substance":"N20"}': [0.2, 0.4, 0.4],
                '{"category":"liku","substance":"N20"}': -1,
                '{"category":"blek","substance":"CH4"}': -1,
            },
        ),
        (
            "Two files, no conflict",
            emiproc.FILES_DIR / "test/profiles/multiple",
            5,
            {
                '{"category":"blek","substance":"CO2"}': [0.3, 0.4, 0.3],
                '{"category":"liku","substance":"CO2"}': [0.5, 0, 0.5],
                '{"category":"liku","substance":"CH4"}': [0.5, 0.1, 0.4],
                '{"category":"blek","substance":"N20"}': [0.2, 0.5, 0.3],
                '{"category":"liku","substance":"N20"}': [0.5, 0.2, 0.3],
                '{"category":"blek","substance":"CH4"}': -1,
            },
        ),
        (
            "Merging points and area sources",
            emiproc.FILES_DIR / "test/profiles/area_vs_point",
            7,
            {
                '{"category":"blek","substance":"CO2", "type":"gridded"}': [
                    0.3,
                    0.4,
                    0.3,
                ],
                '{"category":"liku","substance":"CO2", "type":"gridded"}': [
                    0.5,
                    0,
                    0.5,
                ],
                '{"category":"liku","substance":"CH4", "type":"gridded"}': [
                    0.5,
                    0.1,
                    0.4,
                ],
                '{"category":"blek","substance":"CH4", "type":"gridded"}': [
                    0.2,
                    0.5,
                    0.3,
                ],
                '{"category":"blek","substance":"CO2", "type":"shapped"}': [
                    0.1,
                    0.4,
                    0.5,
                ],
                '{"category":"liku","substance":"CO2", "type":"shapped"}': [
                    0.2,
                    0.7,
                    0.1,
                ],
                '{"category":"liku","substance":"CH4", "type":"shapped"}': [
                    0.3,
                    0.6,
                    0.1,
                ],
                '{"category":"blek","substance":"CH4", "type":"shapped"}': -1,
            },
        ),
        (
            "One file speciated, one file not",
            emiproc.FILES_DIR / "test/profiles/multiple_with_specification",
            4,
            {
                '{"category": "blek", "substance": "CO2"}': [0.4, 0.4, 0.2],
                '{"category": "blek", "substance": "N20"}': [0.1, 0.2, 0.7],
                # These are the same
                '{"category": "liku", "substance": "CO2"}': [0.5, 0.2, 0.3],
                '{"category": "liku", "substance": "N20"}': [0.5, 0.2, 0.3],
            },
        ),
    ],
)
def test_read_v_profiles(name, profiles_dir, n_profiles, expected_dict):
    profiles, indexes = read_vertical_profiles(profiles_dir)

    if profiles is None:
        raise AssertionError(f"Read vertical profile of test case '{name}' failed")

    # Test the values are correct
    assert len(profiles) == n_profiles
    for accessor, expected in expected_dict.items():
        index = indexes.loc[json.loads(accessor)]
        # Missing profile
        if index == -1:
            if expected == -1:
                continue
            else:
                raise AssertionError(
                    f"Read vertical profile of test case '{name}' failed\nExpected"
                    f" {expected} for {accessor=}, got {index=}"
                )
        received = profiles[index].ratios
        if not all(received == np.array(expected)):
            raise AssertionError(
                f"Read vertical profile of test case '{name}' failed\nExpected"
                f" {expected} for {accessor=}, got {received}"
            )


@pytest.mark.parametrize(
    "name,profiles_dir,n_profiles,expected_dict",
    [
        (
            "Simple file",
            emiproc.FILES_DIR / "test/profiles/simple_temporal/test_file.csv",
            4,
            {
                '{"category":"blek","substance":"CO2"}': [
                    WeeklyProfile([0.1, 0.2, 0.3, 0.4, 0.0, 0.0, 0.0])
                ],
                '{"category":"liku","substance":"CO2"}': [
                    WeeklyProfile([0.1, 0.3, 0.2, 0.4, 0.0, 0.0, 0.0])
                ],
                '{"category":"liku","substance":"CH4"}': [
                    WeeklyProfile([0.1, 0.4, 0.2, 0.3, 0.0, 0.0, 0.0])
                ],
                '{"category":"blek","substance":"N20"}': [
                    WeeklyProfile([0.1, 0.5, 0.2, 0.2, 0.0, 0.0, 0.0])
                ],
                '{"category":"liku","substance":"N20"}': -1,
                '{"category":"blek","substance":"CH4"}': -1,
            },
        ),
    ],
)
def test_t_profiles(name, profiles_dir, n_profiles, expected_dict):
    profiles, indexes = read_temporal_profiles(profiles_dir)

    if profiles is None:
        raise AssertionError(f"Read temporal profile of test case '{name}' failed")

    # Test the values are correct
    assert len(profiles) == n_profiles

    for accessor, expected in expected_dict.items():
        index = indexes.loc[json.loads(accessor)]
        # Missing profile
        if index == -1:
            if expected == -1:
                continue
            else:
                raise AssertionError(
                    f"Read temporal profile of test case '{name}' failed\nExpected"
                    f" {expected} for {accessor=}, got {index=}"
                )
        received = profiles[index]
        for r in received:
            assert isinstance(r, TemporalProfile)
        for e in expected:
            assert isinstance(e, TemporalProfile)

        if len(received) != len(expected):
            raise AssertionError(
                f"Read temporal profile of test case '{name}' failed\nExpected"
                f" {expected} for {accessor=}, got {received}"
            )

        for r in received:
            t = type(r)
            # Find the expected profile
            possilbe = [e for e in expected if isinstance(e, t)]
            if len(possilbe) != 1:
                raise AssertionError(
                    f"Read temporal profile of test case '{name}' failed\nExpected"
                    f" {expected} for {accessor=}, got {received}"
                )
            assert (
                r == possilbe[0]
            ), f" Temporal profile do not match received {r}, expected {possilbe[0]}"


def test_no_files():
    proffile = emiproc.FILES_DIR / "test/profiles/simple_vertical_doesnot_exist/"

    with pytest.raises(FileNotFoundError):
        profiles, indexes = read_vertical_profiles(proffile)


if __name__ == "__main__":
    # Test only this file
    pytest.main([__file__])



import geopandas as gpd
import numpy as np
import pandas as pd

from emiproc.tests_utils.test_grids import regular_grid_africa
from emiproc.inventories import Inventory
from emiproc.utilities import get_country_mask

gdf = gpd.GeoDataFrame(
    data={
        ("test", "CO2"): np.random.rand(len(regular_grid_africa.gdf)),
        ("test", "NH3"): np.random.rand(len(regular_grid_africa.gdf)),
        ("test2", "CO2"): np.random.rand(len(regular_grid_africa.gdf)),
    },
    geometry=regular_grid_africa.gdf.geometry,
)


african_inv = Inventory.from_gdf(gdf)

# Now we make an invenotry but only the land cells have emissons (the ones that have a country in them)
da_african_ratios = get_country_mask(regular_grid_africa, return_fractions=True)
mask_cell_no_country = (da_african_ratios.sum(dim="country") == 0).to_numpy()

gdf_land_only = gdf.copy(deep=True)
# Set 0 emissions in the ocean
gdf_land_only.loc[
    mask_cell_no_country, [col for col in gdf.columns if gdf[col].dtype == float]
] = 0.0
african_inv_emissions_only_land = Inventory.from_gdf(gdf_land_only)



import geopandas as gpd
import pytest
import xarray as xr

from emiproc.grids import Grid
from emiproc.inventories import Inventory
from emiproc.profiles.vertical_profiles import VerticalProfiles
from emiproc.tests_utils.test_inventories import inv, inv_with_pnt_sources


@pytest.mark.parametrize(
    "inv",
    (
        inv,
        inv_with_pnt_sources,
    ),
)
def test_correct_fromat(inv: Inventory):
    """Test that the inventory is in the correct format"""

    assert hasattr(inv, "grid")
    if inv.grid is not None:
        assert isinstance(inv.grid, Grid)

    # Gdf or gdfs must be specified
    assert not (inv.gdf is None and inv.gdfs is None)
    if inv.gdf is not None:
        assert isinstance(inv.gdf, gpd.GeoDataFrame)
    if inv.gdfs is not None:
        assert isinstance(inv.gdfs, dict)
        for gdf in inv.gdfs.values():
            assert isinstance(gdf, gpd.GeoDataFrame)

    if hasattr(inv, "v_profiles") and inv.v_profiles is not None:
        assert isinstance(inv.v_profiles, VerticalProfiles)
        assert hasattr(inv, "v_profiles_indexes")
        assert isinstance(inv.v_profiles_indexes, xr.DataArray)



"""File containing the TNO inventory functions."""

import logging
from os import PathLike
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from shapely.creation import polygons

from emiproc.grids import WGS84, GeoPandasGrid, TNOGrid
from emiproc.inventories import Inventory
from emiproc.profiles import naming
from emiproc.profiles.operators import group_profiles_indexes
from emiproc.profiles.temporal_profiles import (
    CompositeTemporalProfiles,
    DayOfYearProfile,
    get_leap_year_or_normal,
    read_temporal_profiles,
)
from emiproc.profiles.utils import ratios_dataarray_to_profiles
from emiproc.profiles.vertical_profiles import read_vertical_profiles

logger = logging.getLogger(__name__)


class TNO_Inventory(Inventory):
    """The TNO inventory.

    TNO has grid cell sources and point sources.
    This handles both.

    https://topas.tno.nl/emissions/


    All the information of the inventory is stored in a netcdf file.

    Each substance is a separate variable in the netcdf file.
    This class will read the `long_name` attribute of each variable to
    determine the substance variables. The `long_name` attribute should
    start with `emission of` .
    You can then merge the substances from the file to a new set of
    substances using the `substances_mapping` argument.
    A default mapping which should work for general cases is provided.

    In the profile files, if you specifiy the substances, you will have to
    use the names created by the mapping, not the names in the nc file.


    :param tno_ds: The xarray dataset with the TNO emission data.
    """

    grid: TNOGrid
    tno_ds: xr.Dataset

    def __init__(
        self,
        nc_file: PathLike,
        substances_mapping: dict[str, str] = {
            "co2_ff": "CO2",
            "co2_bf": "CO2",
            "co_ff": "CO",
            "co_bf": "CO",
            "nox": "NOx",
            "ch4": "CH4",
            "nmvoc": "VOC",
        },
        profiles_dir: PathLike = None,
        vertical_profiles_dir: PathLike = None,
        temporal_profiles_dir: PathLike = None,
        # I assume it is (no info in nc file)
        crs: str = WGS84,
    ) -> None:
        """Create a TNO_Inventory.

        :arg nc_file: Path to the TNO NetCDF dataset.
        :arg substances_mapping: How to map the names from the nc files
            to names for emiproc. See in :py:class:`TNO_Inventory` for more
            information.
            Specifiying the mapping is important if the names in the nc file
            are not the same as the ones you want to use in emiproc.
        :arg profiles_dir: The directory where the profiles are stored.
            If None the same directory as the nc_file is used.
        :arg vertical_profiles_dir: The directory where the vertical profiles
            are stored. If None profiles_dir is used.
        :arg temporal_profiles_dir: The directory where the temporal profiles
            are stored. If None profiles_dir is used.
        """
        super().__init__()

        nc_file = Path(nc_file)
        if not nc_file.is_file():
            raise FileNotFoundError(f"TNO Inventory file {nc_file} is not a file.")

        if profiles_dir is None:
            profiles_dir = nc_file.parent
        else:
            profiles_dir = Path(profiles_dir)
            if not profiles_dir.is_dir():
                raise FileNotFoundError(
                    f"Profiles directory {profiles_dir} is not a directory."
                )

        if vertical_profiles_dir is None:
            vertical_profiles_dir = profiles_dir
        else:
            vertical_profiles_dir = Path(vertical_profiles_dir)

        if temporal_profiles_dir is None:
            temporal_profiles_dir = profiles_dir
        else:
            temporal_profiles_dir = Path(temporal_profiles_dir)
            if not temporal_profiles_dir.is_dir():
                raise FileNotFoundError(
                    f"Temporal profiles directory {temporal_profiles_dir} is not a"
                    " directory."
                )

        # Read the vertical profiles files
        v_profiles, v_profiles_indexes = read_vertical_profiles(vertical_profiles_dir)

        # Set the Temporal profiles
        # Time profiles can vary on category and also on substance
        t_profiles, t_profiles_indexes = read_temporal_profiles(
            temporal_profiles_dir,
            profile_csv_kwargs={
                "encoding": "latin",
            },
        )
        logging.debug(f"Temporal profiles indexes: {t_profiles_indexes}")

        self.name = nc_file.stem

        ds = xr.load_dataset(nc_file, engine="netcdf4")
        self.tno_ds = ds

        self.grid = TNOGrid(nc_file)

        # Read the source types codes
        source_types = ds["source_type_code"].to_numpy()
        for i, source_type in enumerate(source_types):
            # Indexes start at 1 🤦‍♀️
            mask = ds["source_type_index"] == i + 1
            if source_type == b"a":
                mask_area_sources = mask
            elif source_type == b"p":
                mask_point_sources = mask
            else:
                raise NotImplementedError(f"Unknown `source_type_code` {source_type}.")
        # Check that we got all the sources
        masks = [mask_area_sources, mask_point_sources]
        if not all(np.logical_xor.reduce(masks)):
            raise ValueError(
                "The masks overlap or not all point sources were assigned. This is"
                " probably a problem with how the source types are defined or assigned."
            )
        categories = ds["emis_cat_code"].to_numpy()
        # Decode from bytes
        categories = [c.decode("utf-8") for c in categories]

        # Read the variables in the file
        file_substances = [
            var
            for var, xr_var in ds.variables.items()
            if xr_var.attrs.get("long_name", "").startswith("emission of")
        ]
        if substances_mapping:
            # Check that all the file substances are in the mapping
            missing_substances = set(file_substances) - set(substances_mapping.keys())
            if missing_substances:
                self.logger.warning(
                    f"Substances {missing_substances} in the nc file are not in the"
                    f" mapping: {substances_mapping}.\n"
                    "They will be ignored."
                )
                file_substances = [
                    s for s in file_substances if s not in missing_substances
                ]
            # Check if all the mapping substances are in the file
            missing_substances = set(substances_mapping.keys()) - set(file_substances)
            if missing_substances:
                raise ValueError(
                    f"Substances {missing_substances} in the mapping are not in the"
                    f" nc file: {file_substances}.\n"
                    "Please check the names in the mapping."
                )
        else:
            substances_mapping = {s: s for s in file_substances}

        weights = xr.DataArray(
            data=np.zeros((len(categories), len(substances_mapping))),
            coords={
                "category": categories,
                "substance": list(substances_mapping.keys()),
            },
            dims=["category", "substance"],
        )

        polys = self.grid.cells_as_polylist

        # Index in the polygon list (from the gdf) (index start at 1 )
        poly_ind = (
            (ds["longitude_index"] - 1) * self.grid.ny + (ds["latitude_index"] - 1)
        ).to_numpy()
        mapping = {}
        self.gdfs = {}

        for cat_idx, cat_name in enumerate(categories):
            # Indexes start at 1
            mask_this_category = ds["emission_category_index"] == cat_idx + 1

            # Extract point sources information
            mask_points_this_cat = mask_this_category & mask_point_sources
            point_sources_gdf = gpd.GeoDataFrame(
                geometry=gpd.GeoSeries.from_xy(
                    ds["longitude_source"][mask_points_this_cat],
                    ds["latitude_source"][mask_points_this_cat],
                ),
                crs=crs,
            )
            # Extract the emissions for the points sources of this cat
            ds_point_sources = ds[substances_mapping.keys()].sel(
                {"source": mask_points_this_cat}
            )

            for sub_in_nc, sub_emiproc in substances_mapping.items():
                tuple_idx = (cat_name, sub_emiproc)
                if tuple_idx not in mapping:
                    mapping[tuple_idx] = np.zeros(len(polys))
                # Add all the area sources corresponding to that category
                mask = mask_this_category & mask_area_sources
                emissions = ds[sub_in_nc][mask].to_numpy()
                np.add.at(mapping[tuple_idx], poly_ind[mask], emissions)
                weights.loc[dict(category=cat_name, substance=sub_in_nc)] += np.sum(
                    emissions
                )

                # Add the point sources
                point_sources_values = ds_point_sources[sub_in_nc].to_numpy()
                if sub_emiproc in point_sources_gdf:
                    point_sources_gdf[sub_emiproc] += point_sources_values
                else:
                    point_sources_gdf[sub_emiproc] = point_sources_values
                weights.loc[dict(category=cat_name, substance=sub_in_nc)] += np.sum(
                    point_sources_values
                )

            if len(point_sources_gdf):
                # Add only the non emtpy
                self.gdfs[cat_name] = point_sources_gdf

        self.gdf = gpd.GeoDataFrame(
            mapping,
            geometry=polys,
            crs=crs,
        )

        self.cell_areas = ds["area"].T.to_numpy().reshape(-1)

        for profile_type, profiles_indexes in {
            "vertical": v_profiles_indexes,
            "temporal": t_profiles_indexes,
        }.items():
            if profiles_indexes is None:
                continue
            if "substance" in profiles_indexes.dims:
                # Check that the substances in the profiles match the ones in the
                # nc file
                for sub in profiles_indexes["substance"].data:
                    if sub not in substances_mapping.keys():
                        self.logger.warning(
                            f"Substance {sub} in {profile_type} profiles is not in the"
                            f" nc file ({file_substances}). Please check the names in"
                            " the profile files."
                        )
            if "category" in profiles_indexes.dims:
                # Check that the categories in the profiles match the ones in the
                # nc file
                for cat in profiles_indexes["category"].data:
                    if cat not in categories:
                        self.logger.warning(
                            f"Category {cat} in {profile_type} profiles is not in the"
                            f" nc file {categories}. Please check the names in the"
                            " profile files."
                        )

        # We are going to apply the mapping of substances to the profiles

        groupp_mapping = {}
        for sub_from, sub_to in substances_mapping.items():
            if sub_to not in groupp_mapping:
                groupp_mapping[sub_to] = []
            groupp_mapping[sub_to].append(sub_from)
        if t_profiles is not None:
            if "substance" in t_profiles_indexes.dims:
                (
                    t_profiles,
                    t_profiles_indexes,
                ) = group_profiles_indexes(
                    t_profiles,
                    t_profiles_indexes,
                    weights,
                    groupp_mapping,
                    groupping_dimension="substance",
                )
            self.set_profiles(t_profiles, t_profiles_indexes)
        if v_profiles is not None:
            if "substance" in v_profiles_indexes.dims:
                v_profiles, v_profiles_indexes = group_profiles_indexes(
                    v_profiles,
                    v_profiles_indexes,
                    weights,
                    groupp_mapping,
                    groupping_dimension="substance",
                )
            self.set_profiles(v_profiles, v_profiles_indexes)


def read_tno_gridded_profiles(
    file_path: Path, year: int
) -> tuple[CompositeTemporalProfiles, xr.DataArray, GeoPandasGrid]:
    """Read the TNO gridded profiles.

    :arg file_path: The path to the file.
    :arg year: The year for which to read the profiles.

    :returns: A tuple with the profiles, the indexes and the grid.

        * Profiles: The profiles from the file
        * Indexes: The indexes of the profiles
        * Grid: The grid on which the profiles are defined. Usually not the same
            as the inventory grid.

    """

    df = pd.read_csv(file_path, sep=",")
    df = df.loc[df["year"] == year]

    # We need to create a stacked dataset to extract the profiles and indices
    ds_profiles = df.to_xarray()
    da_profiles = (
        ds_profiles.set_coords(["latitude", "longitude", "POLL", "day", "GNFR"])
        .drop_vars(["year"])["Factor"]
        .rename({"POLL": "substance", "GNFR": "category", "day": "day_of_year"})
    )

    coords_multi = ["category", "substance", "latitude", "longitude", "day_of_year"]
    multi_index = pd.MultiIndex.from_arrays(
        [da_profiles[var].values for var in coords_multi], names=coords_multi
    )

    da_profiles = da_profiles.assign_coords(index=multi_index)
    da_profiles = da_profiles.drop_duplicates("index").unstack()
    da_profiles_stacked_cell = da_profiles.stack(cell=["latitude", "longitude"])

    # Get a grid based on the cell dimension
    def get_delta_over_2(da, dim):
        values = da[dim].values
        diffs = np.diff(values)
        # Check that the diffs are the same
        if not np.allclose(diffs, diffs[0]):
            raise ValueError(f"Differences in {dim} are not the same.")
        return diffs[0] / 2.0

    dx = get_delta_over_2(da_profiles, "longitude")
    dy = get_delta_over_2(da_profiles, "latitude")

    x_coords = da_profiles_stacked_cell.longitude.values
    y_coords = da_profiles_stacked_cell.latitude.values

    coords = np.array(
        [
            [x, y]
            for x, y in zip(
                [x_coords - dx, x_coords - dx, x_coords + dx, x_coords + dx],
                [y_coords - dy, y_coords + dy, y_coords + dy, y_coords - dy],
            )
        ]
    )

    coords = np.rollaxis(coords, -1, 0)
    gdf = gpd.GeoDataFrame(geometry=polygons(coords), crs="EPSG:4326")

    grid = GeoPandasGrid(gdf, name="TNO_gridded_profiles")

    da_scaling_factors = da_profiles_stacked_cell.drop_vars(
        ["latitude", "longitude"]
    ).rename(day_of_year="ratio")
    sfs, indexes = ratios_dataarray_to_profiles(da_scaling_factors)

    profiles = CompositeTemporalProfiles.from_ratios(
        ratios=sfs / sfs.sum(axis=1).reshape(-1, 1),
        types=[get_leap_year_or_normal(DayOfYearProfile, year=year)],
    )

    return profiles, indexes, grid



from __future__ import annotations

import logging
from datetime import datetime
from os import PathLike
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from emiproc import PROCESS
from emiproc.exports.netcdf import NetcdfAttributes
from emiproc.grids import RegularGrid
from emiproc.inventories import Inventory
from emiproc.profiles.temporal_profiles import create_scaling_factors_time_serie
from emiproc.profiles.utils import get_desired_profile_index
from emiproc.regrid import remap_inventory
from emiproc.utilities import HOUR_PER_YR, PER_M2_UNITS, SEC_PER_YR, Units

logger = logging.getLogger(__name__)


def export_hourly_emissions(
    inv: Inventory,
    path: PathLike,
    start_time: datetime,
    end_time: datetime,
    netcdf_attributes: NetcdfAttributes,
    var_name_format: str = "{substance}_{category}",
    filename_format: str = "%Y%m%dT%H%M%SZ.nc",
    unit: Units = Units.KG_PER_HOUR,
) -> Path:
    """Export the inventory to hourly netcdf files.

    Supports structured and unstructured grids. Supports gridded emissions
    and point sources.

    A file given at a specific hour is valid for the whole hour.
    (ex file for 14h00 is valid from 14h00 to 14h59m59s)

    :param inv: the inventory to export
    :param path: the path to the output directory
    :param netcdf_attributes: NetCDF attributes to add to the file.
        These can be generated using
        :py:func:`emiproc.exports.netcdf.nc_cf_attributes` .
    :param weights_path: Optionally,
        The path to the weights file to use for regridding.
        If not given, the weights will be calculated on the fly.
    :param var_name_format: The format string to use for the variable names.
        The format string should contain two named fields: ``substance`` and ``category``.
    :param filename_format: The format string to use for the file names.
        The format string should contain fields for date and time.
    :param unit: The unit of the emissions.

    """
    # Check if the inventory is gridded
    if inv.gdfs:
        raise NotImplementedError("Shapped sources are not implemented yet")

    if inv.t_profiles_indexes is None or inv.t_profiles_groups is None:
        raise ValueError(
            "The inventory does not contain temporal profiles required for hourly"
            " exports."
        )
    for invalid_dim in ["type", "country", "cell"]:
        if invalid_dim in inv.t_profiles_indexes:
            raise ValueError(f"Temporal profiles with {invalid_dim=} are not supported")

    grid = inv.grid
    is_regular_grid = isinstance(grid, RegularGrid)
    crs = grid.crs

    # add the history
    netcdf_attributes["emiproc_history"] = str(inv.history)
    netcdf_attributes["projection"] = f"{crs}"

    if unit == Units.KG_PER_YEAR:
        conversion_factor = 1.0
    elif unit == Units.KG_PER_HOUR:
        conversion_factor = 1.0 / HOUR_PER_YR
    elif unit == Units.KG_PER_M2_PER_S:
        conversion_factor = 1.0 / SEC_PER_YR / np.array(grid.cell_areas)
    else:
        raise NotImplementedError(f"Unknown {unit=}")

    # Create the scaling factors for all the time profiles
    reqired_profiles_indexes = np.unique(inv.t_profiles_indexes)
    df_scaling_factors = pd.DataFrame(
        {
            index: create_scaling_factors_time_serie(
                start_time=start_time,
                end_time=end_time,
                profiles=inv.t_profiles_groups[index],
            )
            for index in reqired_profiles_indexes
            if index >= -1
        }
    )

    coords = {
        "substance": inv.substances,
        "category": inv.categories,
        "cell": np.arange(len(grid)),
    }

    if is_regular_grid:
        coords["lat"] = (
            "lat",
            grid.lat_range,
            {
                "standard_name": "latitude",
                "long_name": "latitude",
                "units": "degrees_north",
                "comment": "center_of_cell",
                "bounds": "lat_bnds",
                "projection": f"{grid.crs}",
                "axis": "Y",
            },
        )
        coords["lon"] = (
            "lon",
            grid.lon_range,
            {
                "long_name": "longitude",
                "units": "degrees_east",
                "standard_name": "longitude",
                "comment": "center_of_cell",
                "bounds": "lon_bnds",
                "projection": f"{grid.crs}",
                "axis": "X",
            },
        )

    data_dim = ["lat", "lon"] if is_regular_grid else ["cell"]

    base_ds = xr.Dataset(
        coords=coords,
        attrs=netcdf_attributes,
    )
    if unit in PER_M2_UNITS:
        # add the cell area
        areas = np.array(grid.cell_areas).reshape(grid.shape).T
        if not is_regular_grid:
            areas = areas.reshape(-1)
        base_ds["cell_area"] = (
            data_dim,
            areas,
            {
                "standard_name": "cell_area",
                "long_name": "cell_area",
                "units": "m2",
                "comment": "area of the cell",
                "projection": f"{crs}",
            },
        )
    path = Path(path)
    logger.log(PROCESS, f"Exporting hourly emissions to {path}")

    # Iterrate over time
    for dt, row in df_scaling_factors.iterrows():
        ds = base_ds.copy()
        ds["time"] = dt
        vars = {}
        for cat in inv.categories:
            for sub in inv.substances:
                # Get the scaling factor
                try:
                    index = get_desired_profile_index(
                        inv.t_profiles_indexes, cat=cat, sub=sub
                    )
                except ValueError as ve:
                    logger.warning(
                        f"Could not find profile for {cat=} {sub=}: {ve} \n Assuming"
                        " constant profile"
                    )
                    index = -1

                if index == -1:
                    scaling_factor = 1.0
                else:
                    scaling_factor = row[index]
                if (cat, sub) not in inv.gdf.columns:
                    # Ignore non present cat-sub
                    continue
                # Get the emissions
                emissions = inv.gdf[(cat, sub)].to_numpy().astype(float)
                # Multiply by the scaling factor
                emissions *= scaling_factor * conversion_factor
                name = var_name_format.format(substance=sub, category=cat)

                if is_regular_grid:
                    emissions = emissions.reshape(grid.shape).T

                vars[name] = xr.DataArray(
                    emissions,
                    dims=data_dim,
                    attrs={
                        "standard_name": f"{sub}_{cat}",
                        "long_name": f"{sub}_{cat}",
                        "units": str(unit.value),
                        "comment": f"emissions of {sub} in {cat}",
                    },
                    name=name,
                )
        # Add to the dataset
        ds.update(vars)
        dt: pd.Timestamp

        ds.to_netcdf(path / f"{dt.strftime(filename_format)}")



"""Maps the swiss inventory to Icon.

This example is for a simulation over the city of zuirch.
However the domains goes far beyond the city limits, and even touches germany.

So here we will do a triple nesting:
- We put the zurich inventory in the city limits
- We put the swiss inventory in the country limits but we remove the zurich part
- We put the tno inventory outside the country limits

"""

# %% Imports
from datetime import datetime
from pathlib import Path
from emiproc.inventories.zurich import MapLuftZurich
from emiproc.inventories.tno import TNO_Inventory
from emiproc.inventories.swiss import (
    SwissRasters,
    activities_to_categories,
    default_point_source_correction,
    PointSourceCorrection,
)
from emiproc.inventories.utils import (
    add_inventories,
    crop_with_shape,
    group_categories,
    drop,
)
from emiproc.speciation import merge_substances
from emiproc.grids import LV95, WGS84, ICONGrid
from shapely.geometry import Polygon, Point
from emiproc.regrid import remap_inventory, get_weights_mapping

from emiproc.inventories.categories_groups import CH_2_GNFR, TNO_2_GNFR
from emiproc.inventories.zurich.gnrf_groups import ZH_2_GNFR
import geopandas as gpd
from emiproc.exports.icon import TemporalProfilesTypes, export_icon_oem
from emiproc.utilities import get_natural_earth
from emiproc.profiles.temporal_profiles import from_yaml
from emiproc.profiles.vertical_profiles import VerticalProfile
import numpy as np
import xarray as xr

from emiproc import FILES_DIR

# %% Select the path with my data
data_path = Path(r"C:\Users\coli\Documents\ZH-CH-emission\Data\CHEmissionen")
version = f"v{datetime.now().strftime('%y%m%d')}"
year = 2022
# %%
tno_path = Path(
    r"C:\Users\coli\Documents\Data\TNO\TNO_GHGco_6x6km_v5_0_year2021\inventory\CoCO2_v5_0_emissions_year2021.nc"
)
inv_tno = TNO_Inventory(tno_path, substances_mapping={"co2_bf": "CO2", "co2_ff": "CO2"})
groupping_tno = {}
for cat in inv_tno.categories:
    # We need only the first letter for the categories
    gnfr_cat = f"GNFR_{cat[0]}"
    if gnfr_cat not in groupping_tno:
        groupping_tno[gnfr_cat] = []
    groupping_tno[gnfr_cat].append(cat)
inv_tno.to_crs(LV95)

# %% Create the inventory object
# Change the district heating to the A category
activities_to_categories["1.c"] = "districtheatingps"
default_point_source_correction["districtheatingps"] = (
    PointSourceCorrection.IS_ONLY_POINT_SOURCE
)
CH_2_GNFR["GNFR_A"].append("districtheatingps")
inv_ch = SwissRasters(
    filepath_csv_totals=r"C:\Users\coli\Documents\emissions_preprocessing\output\CH_emissions_EMIS-Daten_1990-2050_Submission_2024_CO2_CO2_biog.csv",
    filepath_point_sources=r"C:\Users\coli\Documents\emissions_preprocessing\input\SwissPRTR-Daten_2007-2022.xlsx",
    rasters_dir=data_path / "ekat_gridascii_v_swiss2icon",
    rasters_str_dir=data_path / "ekat_str_gridascii_v_swiss2icon",
    requires_grid=True,
    year=year,
)
inv_ch = drop(inv_ch, categories=["na"])
inv_ch = merge_substances(inv_ch, {"CO2": ["CO2", "CO2_biog"]}, inplace=True)
inv_ch

# %% load mapluft
zh_path = Path(r"C:\Users\coli\Documents\Data\mapluft") / f"mapLuft_{year}_v2024.gdb"
inv_zh = MapLuftZurich(zh_path, substances=["CO2"])


# %% Load the icon grid
grid_file = Path(
    r"C:\Users\coli\Documents\ZH-CH-emission\for_nikolai\icon_Zurich_R19B9_beo_v3_DOM01.nc"
)
weights_path = grid_file.with_suffix("")
weights_path.mkdir(parents=True, exist_ok=True)
icon_grid = ICONGrid(grid_file)


# %%
def load_zurich_shape(
    zh_raw_file=r"C:\Users\coli\Documents\ZH-CH-emission\Data\Zurich_borders.txt",
    crs_file: int = WGS84,
    crs_out: int = LV95,
) -> Polygon:
    with open(zh_raw_file, "r") as f:
        points_list = eval(f.read())
        zh_poly = Polygon(points_list[0])
        zh_poly_df = gpd.GeoDataFrame(geometry=[zh_poly], crs=crs_file).to_crs(crs_out)
        zh_poly = zh_poly_df.geometry.iloc[0]
        return zh_poly


zh_poly = load_zurich_shape()

gdf = get_natural_earth(resolution="10m", category="cultural", name="admin_0_countries")
gdf = gdf.to_crs(LV95)
ch_poly = gdf.set_index("SOVEREIGNT").loc["Switzerland"].geometry

# %% crop using zurich and swiss shapes


cropped_ch = crop_with_shape(inv_ch, zh_poly, keep_outside=True, modify_grid=True)
cropped_zh = crop_with_shape(
    inv_zh,
    zh_poly,
    keep_outside=False,
)
cropped_tno = crop_with_shape(inv_tno, ch_poly, keep_outside=True, modify_grid=True)


# %% group the categories
groupped_ch = group_categories(cropped_ch, CH_2_GNFR, ignore_missing=True)
groupped_zh = group_categories(cropped_zh, ZH_2_GNFR, ignore_missing=True)
groupped_tno = group_categories(
    cropped_tno,
    groupping_tno,
)

# Merge the groups
groups = {
    key: sum(
        [ZH_2_GNFR.get(key, []), CH_2_GNFR.get(key, []), groupping_tno.get(key, [])], []
    )
    for key in (ZH_2_GNFR | CH_2_GNFR | groupping_tno).keys()
}

# %%

remaped_ch = remap_inventory(
    groupped_ch, icon_grid, weights_path / f"remap_chnew_{year}_2icon"
)
remaped_zh = remap_inventory(
    groupped_zh, icon_grid, weights_path / f"remap_zh2icon_{zh_path.stem}"
)
remaped_tno = remap_inventory(
    groupped_tno, icon_grid, weights_path / f"remap_tno2icon_{tno_path.stem}"
)
# %%
combined = add_inventories(remaped_ch, remaped_zh)
combined = add_inventories(remaped_tno, combined)

# %%
groupped_tno.t_profiles_indexes
# %% Add custom profiles


ships_profiles = from_yaml(FILES_DIR / "profiles" / "yamls" / "ship.yaml")

combined.set_profile(
    profile=ships_profiles,
    category="GNFR_G",
)
# R is the others category and not profile is defined in TNO
combined.set_profile(
    profile=[],
    category="GNFR_R",
)
combined.set_profile(
    # Ground emission (1) at 0m (ground level)
    profile=VerticalProfile(ratios=np.array([1.0, 0.0]), height=np.array([1.0, 2.0])),
    category="GNFR_R",
)

# %%

out_dir = (
    grid_file.parent / f"{grid_file.stem}_zh_ch_tno_combined_year_{year}_{version}"
)

for profile in [TemporalProfilesTypes.HOUR_OF_YEAR, TemporalProfilesTypes.THREE_CYCLES]:
    export_icon_oem(
        inv=combined,
        icon_grid_file=grid_file,
        output_dir=out_dir,
        group_dict=groups,
        substances=["CO2"],
        year=year,
        temporal_profiles_type=profile,
    )


# %% Add the fraction in zurich


weights = get_weights_mapping(
    weights_filepath=None,
    # weights_filepath=weights_path / f"remap_zh_boundaries_2_{grid_file.stem}",
    shapes_inv=icon_grid.gdf.to_crs(LV95),
    shapes_out=[zh_poly],
)


ds_icon = xr.load_dataset(out_dir / "oem_gridded_emissions.nc")
fraction = np.zeros(ds_icon["cell"].shape)
fraction[weights["inv_indexes"]] = weights["weights"]
ds_icon["fraction_in_zurich"] = (
    "cell",
    fraction,
    {
        "units": "-",
        "long_name": "Fraction of the cell in Zurich boundaries",
        "created_by_emiproc": f"{Path(__file__).name}",
    },
)
ds_icon.to_netcdf(out_dir / "oem_gridded_emissions.nc")
ds_icon
# %%



import numpy as np

from emiproc.utilities import get_timezone_mask
from emiproc.grids import RegularGrid


def test_create_simple_mask():
    grid = RegularGrid(
        xmin=47.5,
        xmax=58.5,
        ymin=7.5,
        ymax=12.5,
        nx=13,
        ny=10,
    )
    arr = get_timezone_mask(grid)
    # Check that the array is a numpy array of strings
    assert isinstance(arr, np.ndarray)
    # We don't know the lengths of the strings
    assert arr.dtype.kind == "U"

    # Check the sized is the same as the grid
    assert arr.shape == grid.shape



# %%
import pytest
import geopandas as gpd
from emiproc.grids import RegularGrid
from emiproc.tests_utils.test_grids import regular_grid


# %%
def test_creation():
    RegularGrid(
        xmin=-1, xmax=5, ymin=-2, ymax=3, nx=10, ny=15, name="Test Regular Grid"
    )


def test_creation_n_d():
    RegularGrid(
        xmin=-1, ymin=-2, dx=0.1, dy=0.2, nx=10, ny=15, name="Test Regular Grid"
    )


def test_creation_max_d():
    RegularGrid(
        xmin=-1, ymin=-2, xmax=5, ymax=3, dx=0.1, dy=0.2, name="Test Regular Grid"
    )

def test_creation_fails_max_d_n():
    pytest.raises(ValueError, RegularGrid,
        xmin=-1, ymin=-2, xmax=5, ymax=3, dx=0.1,dy=0.2, nx=10,ny=15, name="Test Regular Grid"
    )
def test_creation_fails_not_enough():
    pytest.raises(ValueError, RegularGrid,
        xmin=-1, ymin=-2, xmax=5, ymax=3,  name="Test Regular Grid"
    )
    pytest.raises(ValueError, RegularGrid,
        xmin=-1, ymin=-2, dx=0.1,dy=0.2, name="Test Regular Grid"
    )
    pytest.raises(ValueError, RegularGrid,
        xmin=-1, ymin=-2,  nx=10,ny=15,  name="Test Regular Grid"
    )

def test_polylist():
    regular_grid.cells_as_polylist

    # Test how we iterate over the cells in the polygon list


def test_area():
    regular_grid.cell_areas


def test_shape():
    nx, ny = regular_grid.shape

    assert nx == regular_grid.nx
    assert ny == regular_grid.ny



import pytest

import geopandas as gpd

from emiproc import TESTS_DIR
from emiproc.exports.icon import export_icon_oem
from emiproc.grids import ICONGrid
from emiproc.inventories import Inventory
from emiproc.regrid import remap_inventory
from emiproc.tests_utils.icon import SIMPLE_ICON_GRID_PATH, get_test_grid
from emiproc.tests_utils.test_inventories import inv_with_pnt_sources


def test_get_test_grid():
    """Test the function to get the test grid."""

    grid = get_test_grid()
    assert grid is not None


def test_export_icon():
    """Test the export of ICON files."""

    grid = ICONGrid(SIMPLE_ICON_GRID_PATH)

    inv = Inventory.from_gdf(
        gdfs={
            "point": gpd.GeoDataFrame(
                data={
                    "CO2": 3 * [1],
                },
                geometry=gpd.points_from_xy(
                    y=[38.695852, 48.695852, 59.695852],
                    x=[69.089789, 79.089789, 89.089789],
                ),
                crs="WGS84",
            ),
        }
    )

    inv_on_icon = remap_inventory(inv, grid)

    export_icon_oem(
        inv_on_icon, SIMPLE_ICON_GRID_PATH, output_dir=TESTS_DIR / "export_icon"
    )


def test_export_wrong_grid():
    with pytest.raises(ValueError):
        export_icon_oem(
            inv_with_pnt_sources,
            SIMPLE_ICON_GRID_PATH,
            output_dir=TESTS_DIR / "export_icon",
        )


if __name__ == "__main__":
    pytest.main(["-x", __file__])



import pytest
from emiproc.tests_utils.profiles import (
    da_profiles_indexes_catsub,
    da_profiles_indexes_sub,
)

from emiproc.profiles.utils import get_desired_profile_index
from emiproc.profiles.temporal_profiles import SpecificDay, days_of_specific_day


@pytest.mark.parametrize(
    "specific_days, expected",
    (
        (SpecificDay.MONDAY, [SpecificDay.MONDAY]),
        (SpecificDay.TUESDAY, [SpecificDay.TUESDAY]),
        (SpecificDay.WEDNESDAY, [SpecificDay.WEDNESDAY]),
        (SpecificDay.THURSDAY, [SpecificDay.THURSDAY]),
        (SpecificDay.FRIDAY, [SpecificDay.FRIDAY]),
        (SpecificDay.SATURDAY, [SpecificDay.SATURDAY]),
        (SpecificDay.SUNDAY, [SpecificDay.SUNDAY]),
        (SpecificDay.WEEKEND, [SpecificDay.SATURDAY, SpecificDay.SUNDAY]),
        (
            SpecificDay.WEEKDAY,
            [
                SpecificDay.MONDAY,
                SpecificDay.TUESDAY,
                SpecificDay.WEDNESDAY,
                SpecificDay.THURSDAY,
                SpecificDay.FRIDAY,
            ],
        ),
    ),
)
def test_days_of_specific_day(specific_days, expected):
    assert days_of_specific_day(specific_days) == expected


def test_get_desired_profile_index_working_cases():
    assert (
        get_desired_profile_index(da_profiles_indexes_catsub, cat="b", sub="CO2") == 2
    )
    assert (
        get_desired_profile_index(da_profiles_indexes_catsub, cat="a", sub="CH4") == 1
    )
    # Test when you add too specific it still works
    assert get_desired_profile_index(da_profiles_indexes_sub, cat="a", sub="CH4") == 1
    assert (
        get_desired_profile_index(
            da_profiles_indexes_catsub, cell=43, cat="a", sub="CH4"
        )
        == 1
    )


def test_get_desired_profile_index_errors():
    # Not specified enough
    pytest.raises(
        ValueError, get_desired_profile_index, da_profiles_indexes_catsub, cat="a"
    )


def test_get_desired_profile_index_wrong_catsub():
    # Category does not exist
    pytest.raises(
        ValueError,
        get_desired_profile_index,
        da_profiles_indexes_catsub,
        cat="CH4",
        sub="CH4",
    )


if __name__ == "__main__":
    pytest.main([__file__])



from emiproc.profiles.temporal_profiles import (
    CompositeTemporalProfiles,
    HourOfLeapYearProfile,
    HourOfYearProfile,
    interpolate_profiles_hour_of_year,
)
from emiproc.tests_utils import temporal_profiles


def test_interpolate_day():

    out_profiles = interpolate_profiles_hour_of_year(
        CompositeTemporalProfiles([[temporal_profiles.daily_test_profile]]),
        year=2022,
        return_profiles=True,
    )

    assert isinstance(out_profiles, CompositeTemporalProfiles)
    assert len(out_profiles.types) == 1
    assert out_profiles.types[0] == HourOfYearProfile


def test_leap_year():

    out_profiles = interpolate_profiles_hour_of_year(
        CompositeTemporalProfiles([[temporal_profiles.daily_test_profile]]),
        year=2024,
        return_profiles=True,
    )

    assert isinstance(out_profiles, CompositeTemporalProfiles)
    assert len(out_profiles.types) == 1
    assert out_profiles.types[0] == HourOfLeapYearProfile


def test_from_day_of_year():
    out_profiles = interpolate_profiles_hour_of_year(
        CompositeTemporalProfiles([[temporal_profiles.day_of_year_test_profile]]),
        year=2022,
        return_profiles=True,
    )

    assert isinstance(out_profiles, CompositeTemporalProfiles)
    assert len(out_profiles.types) == 1
    assert out_profiles.types[0] == HourOfYearProfile


def test_from_day_of_year_after_leap():
    out_profiles = interpolate_profiles_hour_of_year(
        CompositeTemporalProfiles([[temporal_profiles.day_of_year_test_profile]]),
        year=2021,
        return_profiles=True,
    )

    assert isinstance(out_profiles, CompositeTemporalProfiles)
    assert len(out_profiles.types) == 1
    assert out_profiles.types[0] == HourOfYearProfile



from __future__ import annotations
import numpy as np
import pandas as pd
import xarray as xr

from emiproc.inventories import Inventory
from emiproc.profiles.temporal_profiles import get_index_in_profile
from emiproc.utils.translators import inv_to_xarray


def get_temporally_scaled_array(
    inv: Inventory,
    time_range: pd.DatetimeIndex | int,
    sum_over_cells: bool = True,
) -> xr.DataArray:
    """Transform the inventory to a temporally resolved emissions array.

    Missing profiles are assumed to be constant profiles.

    You can easily plot the profiles of you different category and substances
    of the received inventory with the following code:
    ```
    da.sum("cell").stack(catsub=["category", "substance"]).plot.line(x="time")
    ```

    :param inv: the inventory to transform
    :param time_range: the time range to use for the temporal resolution.
        If an integer is given the time range will a daily range of the given year.
    :param sum_over_cells: if True the emissions are summed over the cells.
        This can be useful to improve the performance of the plotting.

    :return: the temporally resolved emissions array.
        The units are the same as in the inventory. (kg/y/cell)
        But now even scaled on the time axis given units are still kg/y/cell.
        If you want to get the emissions at your time resolution you need divide
        by the number of your time resolution that fits in a year.
    """

    profiles, profiles_indexes = inv.t_profiles_groups, inv.t_profiles_indexes

    if profiles is None or profiles_indexes is None:
        raise ValueError(
            "The inventory does not have temporal profiles."
            "You need to set the profiles to get a temporally resolved emissions array."
        )

    if isinstance(time_range, int):
        time_range = pd.date_range(
            start=f"{time_range}-01-01",
            end=f"{time_range}-12-31",
            freq="D",
            inclusive="both",
        )

    da_totals = inv_to_xarray(inv)
    if sum_over_cells:
        da_totals = da_totals.sum("cell")

    # Acess the scaling factors
    scaling_factors_array = profiles.scaling_factors[profiles_indexes]

    if "cell" in profiles_indexes.dims:
        if sum_over_cells:
            raise ValueError(
                "The scaling factors are defined for individual cells."
                "You need to set sum_over_cells to False"
            )
        # The profiles are usually only given on cells with emissions
        missing_cells = da_totals.cell.loc[~da_totals.cell.isin(profiles_indexes.cell)]
        # Check that the profiles are given for all cells
        assert (
            da_totals.sel(cell=missing_cells).sum().values == 0
        ), "Some cell or emissions with none zero values have missing profiles"

    da_scaling_factors = xr.DataArray(
        scaling_factors_array,
        coords=dict(
            **profiles_indexes.coords,
            profile_index=np.arange(profiles.size),
        ),
    )

    # Get teh index of of the scaling factor for each type of temporal profile
    size_offset = 0
    indices = []

    for profile_type in profiles.types:

        indices.append(get_index_in_profile(profile_type, time_range) + size_offset)

        size_offset += profile_type.size

    indices_to_use = xr.DataArray(
        np.array(indices).T,
        coords=dict(
            time=time_range,
            profile=np.arange(len(profiles.types)),
        ),
        dims=["time", "profile"],
    )

    # Get the proper scaling factors for each index
    scaling_factors_at_times = da_scaling_factors.loc[
        dict(profile_index=indices_to_use)
    ]
    # Merge all the time factors together
    scaling_factor_at_times = scaling_factors_at_times.prod("profile")
    # Set the scaling factors on the missing cells
    scaling_factor_at_times_all_cells = scaling_factor_at_times.reindex(
        da_totals.coords
    ).fillna(1.0)

    # Finally scale the emissions at each time
    temporally_scaled_emissions = da_totals * scaling_factor_at_times_all_cells

    return temporally_scaled_emissions



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import geopandas as gpd
import numpy as np
import xarray as xr

from emiproc.profiles.temporal_profiles import (
    CompositeTemporalProfiles,
    SpecificDayProfile,
    TemporalProfile,
)
from emiproc.profiles.utils import (
    get_objects_of_same_type_from_list,
    profiles_to_scalingfactors_dataarray,
    ratios_dataarray_to_profiles,
)
from emiproc.profiles.vertical_profiles import (
    VerticalProfile,
    VerticalProfiles,
    resample_vertical_profiles,
)
from emiproc.utilities import get_country_mask

if TYPE_CHECKING:
    from emiproc.grids import Grid
    from emiproc.inventories import Inventory

logger = logging.getLogger(__name__)


def concatenate_profiles(
    profiles_list: list[
        VerticalProfiles | list[list[TemporalProfile]] | CompositeTemporalProfiles
    ],
) -> VerticalProfiles | CompositeTemporalProfiles:
    # Check that all profiles are of the same type
    profile_type = type(profiles_list[0])
    if not all([type(p) == profile_type for p in profiles_list]):
        raise TypeError(
            "All profiles must be of the same type, got"
            f" {[type(p) for p in profiles_list]}"
        )
    if profile_type == VerticalProfiles:
        return resample_vertical_profiles(*profiles_list)
    elif profile_type == list:
        reduced_list = sum(profiles_list, [])
        return CompositeTemporalProfiles(reduced_list)
    elif profile_type == CompositeTemporalProfiles:
        return CompositeTemporalProfiles.join(*profiles_list)
    else:
        raise TypeError(f"Unknown profile type {profile_type}")


def weighted_combination(
    profiles: VerticalProfiles | list[TemporalProfile], weights: np.ndarray
) -> VerticalProfile | TemporalProfile:
    """Combine the different profiles according to the specified weights.

    Each ratio of the new profile follow the formula:

    .. math::
        r_{new} = \\frac{\\sum_{i}^n w_i r_i}{\\sum_{i}^n w_i}

    where :math:`r_i` is the ratio of the profile i and :math:`w_i` is the weight of the profile i.
    and :math`n` is the number of profiles.

    :arg profiles: The profiles to combine.
        If temporal, they must be all of the same type.
        If vertical, it must be a :py:class:`VerticalProfiles` object.
    :arg weights: The weights to use for the combination.
        See numpy.average() for more details on the weights.

    :return: The combined profile.
    """

    if len(profiles) != len(weights):
        raise ValueError(
            f"The number of profiles and weights must be the same: {len(profiles)=} !="
            f" {len(weights)=}"
        )

    if len(profiles) == 0:
        raise ValueError("No profiles given.")

    if isinstance(profiles, VerticalProfiles):
        return VerticalProfile(
            np.average(profiles.ratios, axis=0, weights=weights), profiles.height.copy()
        )
    elif isinstance(profiles, list):
        # Time profile, make sure they are all of the same type
        type_profile = type(profiles[0])
        if not all([isinstance(p, type_profile) for p in profiles]):
            raise TypeError(
                "All profiles must be of the same type, got"
                f" {[type(p) for p in profiles]}"
            )
        kwargs = {}
        if isinstance(profiles[0], SpecificDayProfile):
            kwargs["specific_day"] = profiles[0].specific_day
        return type_profile(
            ratios=np.average(
                np.array([p.ratios for p in profiles]), axis=0, weights=weights
            ),
            **kwargs,
        )


def weighted_combination_time(
    profiles: list[list[TemporalProfile]], weights: np.ndarray
) -> list[TemporalProfile]:
    """Combine the different profiles according to the specified weights.

    Same as :py:func:`weighted_combination` but for combinated time profiles.
    """
    out_profiles = []
    weights = weights.copy()
    # First read the types from the first combined profile
    for profile in profiles[0]:
        # Find in each list the profile of that type
        p_to_merge = []
        for i, p_list in enumerate(profiles):
            same_types = get_objects_of_same_type_from_list(profile, p_list)
            if len(same_types) > 1:
                raise ValueError(f"Duplicated profiles of type {type(profile)}")
            elif len(same_types) == 0:
                # No profile of this type don't use it in the combination
                weights[i] = 0
                p_to_merge.append(profile)
            else:
                p_to_merge.append(same_types[0])

        out_profiles.append(weighted_combination(p_to_merge, weights))

    return out_profiles


def combine_profiles(
    profiles: (
        VerticalProfiles | list[list[TemporalProfile]] | CompositeTemporalProfiles
    ),
    profiles_indexes: xr.DataArray,
    dimension: str,
    weights: xr.DataArray,
) -> tuple[VerticalProfiles | CompositeTemporalProfiles, xr.DataArray]:
    """Combine profiles from multidimensional array by reducing over a specified dimension.

    The indexes and the weights but be of the same dimensions.

    :arg profiles: The profiles to use for merging.
    :arg profiles_indexes: The profiles indexes of the data.
    :arg dimension: The dimension along which the combination should be done.
    :arg weights: The weights of the data. In terms of emissions, it means
        the total emission of that data.
        Weights can be obtained through :py:func:`get_weights_of_gdf_profiles`

    :return: the new profiles and the indexes of the combined data
        in these new profiles.
        The indexes array has the dimension of combination removed.
    """

    logger = logging.getLogger(__name__)
    logger.debug(f"Combining {profiles_indexes=} on {dimension=} with {weights=}")
    # assert len(indexes.coords[dimension]) > 1, "Cannot combine over 1 dimension"
    if dimension not in profiles_indexes.dims:
        raise ValueError(
            f"Dimension {dimension} not in {profiles_indexes.dims}, cannot combine"
            " over it."
        )

    # Aligns exisiting coordinates
    weights = weights.broadcast_like(profiles_indexes).fillna(0)
    profiles_indexes = profiles_indexes.broadcast_like(weights).fillna(-1).astype(int)

    # Make sure that where the index is -1, the weights are 0
    mask_missing_profile = profiles_indexes == -1
    weights = xr.where(mask_missing_profile, 0.0, weights)

    if isinstance(profiles, list):
        # Make it composite temporal profiles
        profiles = CompositeTemporalProfiles(profiles)
    # Get the ratios to use depending on the profile type
    if not isinstance(profiles, (VerticalProfiles, CompositeTemporalProfiles)):
        raise TypeError(
            f"Unknown profile type {type(profiles)}, must be VerticalProfiles or"
            " CompositeTemporalProfiles"
        )
    # Access the profiles
    da_sf = profiles_to_scalingfactors_dataarray(profiles, profiles_indexes)

    logger.debug(
        f"Creating averaged profiles with {da_sf=} and" f" {weights=} on {dimension=}"
    )
    # Perform the average on the profiles
    new_scaling_factors = (da_sf * weights).sum(dimension)

    # Recover the profiles and indexes from the new scaling factors

    new_profiles, new_indexes = ratios_dataarray_to_profiles(
        new_scaling_factors.rename({"scaling_factors": "ratio"})
    )

    if isinstance(profiles, VerticalProfiles):
        # Rescale the unique profiles to sum to 1
        new_profiles = new_profiles / new_profiles.sum(axis=1).reshape(-1, 1)
        new_profiles = VerticalProfiles(new_profiles, profiles.height)
    elif isinstance(profiles, CompositeTemporalProfiles):
        new_profiles = CompositeTemporalProfiles.from_ratios(
            ratios=new_profiles,
            types=profiles.types,
            rescale=True,
        )
    else:
        raise TypeError(f"Unknown profile type {type(profiles)}")

    return new_profiles, new_indexes


def get_weights_of_gdf_profiles(
    gdf: gpd.GeoDataFrame, profiles_indexes: xr.DataArray
) -> xr.DataArray:
    """Read the emission values from the gdf to get the weights.

    Weights for missing data will be 0. This makes sure that if later
    group different profiles, the data will the merged profile from
    the other categories.

    :arg gdf: The gdf of an inventory, following the standard definition.
    :arg profiles_indexes: Indexes of the profiles to use.

    :return weights: A xarray containing the weights of the given profiles_indexes.
    """

    weights = xr.zeros_like(profiles_indexes, dtype=float)
    for col in gdf.columns:
        if isinstance(gdf[col].dtype, gpd.array.GeometryDtype):
            continue
        cat, sub = col

        # Case indexes is a subset of the data
        if (
            "category" in profiles_indexes.dims
            and cat not in profiles_indexes.coords["category"]
        ) or (
            "substance" in profiles_indexes.dims
            and sub not in profiles_indexes.coords["substance"]
        ):
            continue

        serie = gdf.loc[:, col]

        col_value = (
            # if depend dont get the total of the serie
            serie.loc[profiles_indexes.coords["cell"].values]
            if "cell" in profiles_indexes.dims
            else sum(serie)
        )
        if "category" in profiles_indexes.dims and "substance" in profiles_indexes.dims:
            weights.loc[dict(category=cat, substance=sub)] = col_value
        elif "category" in profiles_indexes.dims:
            weights.loc[dict(category=cat)] += col_value
        elif "substance" in profiles_indexes.dims:
            weights.loc[dict(substance=sub)] += col_value
        else:
            weights += col_value

    # Where we there is no profile, we should not use the weight
    weights = xr.where(profiles_indexes == -1, 0, weights)

    return weights


def group_profiles_indexes(
    profiles: VerticalProfiles | list[list[TemporalProfile]],
    profiles_indexes: xr.DataArray,
    indexes_weights: xr.DataArray,
    categories_group: dict[str, list[str]],
    groupping_dimension: str = "category",
) -> tuple[VerticalProfiles | list[list[TemporalProfile]], xr.DataArray]:
    """Groups profiles and their indexes according to the given mapping.

    It is possible to group categories or substances or cells by setting
    the groupping_dimension to the corresponding dimension.

    :arg profiles: The profiles to use for merging.
    :arg profiles_indexes: The profiles indexes of the data.
    :arg indexes_weights: The weights of the data. In terms of emissions, it means
        the total emission of that data.
        Weights can be obtained through :py:func:`get_weights_of_gdf_profiles`
    :arg categories_group: A dictionary containing the mapping of the categories.
        The keys are the new categories and the values are the list of categories
        that should be grouped together.
    :arg groupping_dimension: The dimension along which the combination should be done.
        Default is "category".
    """
    if not groupping_dimension in profiles_indexes.dims:
        # if they don't depend on category, we don't need to do anything
        return profiles, profiles_indexes

    groups_indexes_list = []
    merged_profiles = []
    n_profiles = 0
    for group, categories in categories_group.items():
        cats_with_profiles = [
            c for c in categories if c in profiles_indexes.coords[groupping_dimension]
        ]
        group_coord = {groupping_dimension: [group]}
        sel_dict = {groupping_dimension: cats_with_profiles}
        if not cats_with_profiles:
            # No categories to group
            # Add unknown profiles
            group_indexes = profiles_indexes.sum(dim=groupping_dimension).expand_dims(
                group_coord
            )
            # set all the values to -1, no matter what the value is
            groups_indexes_list.append(
                xr.where(
                    group_indexes != -1,
                    -1,
                    group_indexes,
                )
            )

            continue

        # Combine the profiles that have to be grouped
        new_profiles, group_indexes = combine_profiles(
            profiles,
            profiles_indexes.sel(**sel_dict),
            dimension=groupping_dimension,
            weights=indexes_weights.sel(**sel_dict),
        )
        # Set that new coord to the dimension
        group_indexes = group_indexes.expand_dims(group_coord)

        # Offset the indexes for merging with the profiles
        groups_indexes_list.append(
            xr.where(
                group_indexes != -1,
                group_indexes + n_profiles,
                -1,
            )
        )
        n_profiles += len(new_profiles)
        merged_profiles.append(new_profiles)

    if not merged_profiles:
        coord_values = profiles_indexes.coords[groupping_dimension].to_numpy()
        raise ValueError(
            f"No profiles to group with {categories_group=}, on dimension"
            f" {groupping_dimension=} with {coord_values=}"
        )

    if isinstance(merged_profiles[0], VerticalProfiles):
        merged_profiles = sum(
            merged_profiles,
        )
    elif isinstance(merged_profiles[0], CompositeTemporalProfiles):
        merged_profiles = CompositeTemporalProfiles.join(
            *merged_profiles,
        )
    else:
        merged_profiles = sum(merged_profiles, [])

    new_indices = (
        xr.concat(
            groups_indexes_list,
            dim=groupping_dimension,
        )
        .fillna(-1)
        .astype(int)
    )

    return merged_profiles, new_indices


def country_to_cells(
    profiles: (
        VerticalProfiles | list[list[TemporalProfile]] | CompositeTemporalProfiles
    ),
    profiles_indexes: xr.DataArray,
    grid: Grid,
    country_mask_kwargs: dict[str, any] = {},
    ignore_missing_countries: bool = False,
) -> tuple[VerticalProfiles | CompositeTemporalProfiles, xr.DataArray]:
    """Takes profiles given for countries and transform them to cells.

    The input profiles must have a 'country' dimension.
    The output profiles will have a 'cell' dimension.

    country_mask_kwargs are passed to :py:func:`emiproc.utilities.get_country_mask`.
    """

    # First check that the profiles are for countries
    if not isinstance(profiles_indexes, xr.DataArray):
        raise TypeError(
            f"Expected {profiles_indexes=} to be a xr.DataArray, got"
            f" {type(profiles_indexes)}"
        )
    if "country" not in profiles_indexes.dims:
        raise ValueError(
            f"Expected {profiles_indexes=} to have a country dimension, got"
            f" {profiles_indexes.dims}"
        )
    if "cell" in profiles_indexes.dims:
        raise ValueError(
            f"Expected {profiles_indexes=} to not have a cell dimension, got"
            f" {profiles_indexes.dims}"
        )
    if isinstance(profiles, list):
        # Make it composite temporal profiles
        profiles = CompositeTemporalProfiles(profiles)

    # Check that the set makes sense
    if not np.all(profiles_indexes >= -1):
        raise ValueError(
            f"Expected {profiles_indexes=} to have all values >= -1, got"
            f" {profiles_indexes.min()}"
        )
    if not np.all(profiles_indexes < len(profiles)):
        raise ValueError(
            f"Expected {profiles_indexes=} to have all values <= {len(profiles)}, got"
            f" {profiles_indexes.max()}"
        )

    # These will be the weights of the profiles
    countries_fractions: xr.DataArray = get_country_mask(
        grid, return_fractions=True, **country_mask_kwargs
    )

    # Check that the countries are all given
    # All the countries of the grid must be in the profiles
    missing_countries = set(countries_fractions.coords["country"].values) - set(
        profiles_indexes.coords["country"].values
    )

    if missing_countries and not ignore_missing_countries:
        raise ValueError(
            f"Missing countries {missing_countries=} in {profiles_indexes=}. "
            "Please check the profiles or set `ignore_missing_countries=False` ."
        )

    da_sf = profiles_to_scalingfactors_dataarray(profiles, profiles_indexes)

    sf_on_cell = da_sf.dot(countries_fractions, dims=["country"])

    profiles_array, new_indexes = ratios_dataarray_to_profiles(
        sf_on_cell.rename({"scaling_factors": "ratio"})
    )

    if isinstance(profiles, VerticalProfiles):
        new_profiles = VerticalProfiles(
            profiles_array / profiles_array.sum(axis=1).reshape((-1, 1)),
            profiles.height,
        )
    elif isinstance(profiles, CompositeTemporalProfiles):
        new_profiles = CompositeTemporalProfiles.from_ratios(
            profiles_array, types=profiles.types, rescale=True
        )
    else:
        raise TypeError(f"Unknown profile type {type(profiles)}")

    return new_profiles, new_indexes


def remap_profiles(
    profiles: VerticalProfiles | CompositeTemporalProfiles,
    profiles_indexes: xr.DataArray,
    emissions_weights: xr.DataArray,
    weights_mapping: dict[str, np.ndarray],
) -> tuple[VerticalProfiles | CompositeTemporalProfiles, xr.DataArray]:
    """Remap the profiles on a new grid.

    :arg profiles: The profiles to remap.
    :arg profiles_indexes: The indexes of the profiles.
    :arg emissions_weights: The weights of the emissions.
        Can be calculated using :py:func:`get_weights_of_gdf_profiles`.
    :arg weights_mapping: A dictionary containing the weights for the remapping.
        This is the result of :py:func:`emiproc.utilities.get_weights_mapping`.

    :return: The remapped profiles and the new indexes.

    """
    if isinstance(profiles, VerticalProfiles):
        raise NotImplementedError("Vertical profiles remapping not implemented yet.")
    if not isinstance(profiles, CompositeTemporalProfiles):
        raise TypeError(
            f"Invalid profile type {type(profiles)}, must be {CompositeTemporalProfiles}"
        )

    assert "cell" in profiles_indexes.dims, "Expected cell dimension in indexes"
    assert "cell" in emissions_weights.dims, "Expected cell dimension in weights"

    # Check that all the cells are in the profiles
    missing_cells_in_profiles = set(weights_mapping["inv_indexes"]) - set(
        profiles_indexes["cell"].values
    )
    if missing_cells_in_profiles:
        # Get a new weights mapping without the missing cells
        mask_missing = np.isin(
            weights_mapping["inv_indexes"], list(missing_cells_in_profiles)
        )
        weights_mapping = {k: v[~mask_missing] for k, v in weights_mapping.items()}

    # Get the emissions weights at the remapping index
    da_weights_of_remapping = xr.DataArray(
        weights_mapping["weights"],
        dims=["cell"],
        coords={"cell": weights_mapping["inv_indexes"]},
    )
    da_weights = (
        emissions_weights.sel(cell=weights_mapping["inv_indexes"])
        * da_weights_of_remapping
    )

    # Set the profiles needed on the remapping cells
    profiles_to_get_remapped = profiles_indexes.sel(cell=weights_mapping["inv_indexes"])

    # access the profiles to get the ratios at each inpout
    da_ratios = (
        xr.DataArray(
            profiles.ratios[profiles_to_get_remapped],
            dims=[*profiles_to_get_remapped.dims, "ratio"],
            coords={
                **profiles_to_get_remapped.coords,
                "ratio": range(profiles.ratios.shape[1]),
            },
            # Remove the profiles with no ratios (will be set to nan)
            # This assumes that no profile = no contribution, so only the other ratios in the cell will have an impact
        )
        .where(profiles_to_get_remapped != -1)
        .assign_coords(output_cell=("cell", weights_mapping["output_indexes"]))
    )

    # Do the actual remapping calculations
    da_remapped_ratios = (
        (da_ratios * da_weights)
        .groupby("output_cell")
        .mean()
        .rename({"output_cell": "cell"})
    )

    new_ratios, new_indices = ratios_dataarray_to_profiles(da_remapped_ratios)
    new_profiles = CompositeTemporalProfiles.from_ratios(
        new_ratios, profiles.types, rescale=True
    )

    return new_profiles, new_indices


def add_profiles(
    inv1: Inventory,
    inv2: Inventory,
) -> tuple[CompositeTemporalProfiles, xr.DataArray]:
    """Add the profiles of two inventories together.

    :arg inv1: The first inventory.
    :arg inv2: The second inventory.

    :return: The sum of the two profiles.
    """

    indexes_name = "t_profiles_indexes"
    profiles_name = "t_profiles_groups"

    indexes1 = getattr(inv1, indexes_name)
    indexes2 = getattr(inv2, indexes_name)

    # Aligns exisiting coordinates
    indexes1, indexes2 = xr.broadcast(indexes1, indexes2)
    # Add the missing and convert again to integers
    indexes1 = indexes1.fillna(-1).astype(int)
    indexes2 = indexes2.fillna(-1).astype(int)

    weights1 = get_weights_of_gdf_profiles(inv1.gdf, profiles_indexes=indexes1)
    weights2 = get_weights_of_gdf_profiles(inv2.gdf, profiles_indexes=indexes2)
    weights1, weights2 = xr.broadcast(weights1, weights2)
    # Add the missing
    weights1 = weights1.fillna(0)
    weights2 = weights2.fillna(0)

    profiles_name = "t_profiles_groups"
    profiles1: CompositeTemporalProfiles = getattr(inv1, profiles_name)
    profiles2: CompositeTemporalProfiles = getattr(inv2, profiles_name)

    # Make the profiles have the same sub-profiles included
    # This will make scaling factors of 1 when a sub-profile is missing
    all_types = set(sum([p.types for p in [profiles1, profiles2]], []))
    # Careful here, becaue the types will change the order of position
    profiles1 = profiles1.broadcast(all_types)
    profiles2 = profiles2.broadcast(all_types)

    # Make sure they are the same in the new profiles
    assert (
        profiles1.types == profiles2.types
    ), "Profiles not same type. Please raise an issue on Github."

    sf1 = profiles_to_scalingfactors_dataarray(profiles1, indexes1)
    sf2 = profiles_to_scalingfactors_dataarray(profiles2, indexes2)

    # Multiply the scaling factors by the weights
    sf_total = sf1 * weights1 + sf2 * weights2

    ratios, indexes = ratios_dataarray_to_profiles(
        sf_total.rename({"scaling_factors": "ratio"})
    )

    # Create the new profiles object
    profiles = CompositeTemporalProfiles.from_ratios(
        ratios, rescale=True, types=profiles1.types
    )

    return profiles, indexes


def add_constant_profile_to_missing_cells(inv: Inventory) -> Inventory:

    out_inv = inv.copy(profiles=False)

    for profile_name, indexes_name in [
        ("t_profiles_groups", "t_profiles_indexes"),
        ("v_profiles", "v_profiles_indexes"),
    ]:
        profiles = getattr(inv, profile_name)
        indexes = getattr(inv, indexes_name)

        if profiles is None or indexes is None:
            continue

        if "cell" not in indexes.dims:
            # No profile means constant profile
            out_inv.set_profiles(
                profiles,
                indexes=indexes,
            )
            continue

        # Append the constant profile
        if isinstance(profiles, CompositeTemporalProfiles):
            # new_profile = CompositeTemporalProfiles.join(
            #    profiles, CompositeTemporalProfiles([[]])
            # )
            new_profile = profiles
        else:
            raise NotImplementedError(
                "function `add_constant_profile_to_missing_cells` not implemented for "
                f"{type(profiles)}"
            )

        missing_cells = set(np.arange(len(inv.grid))) - set(indexes["cell"].values)

        new_indexes = xr.concat(
            [
                indexes,
                xr.full_like(indexes.isel(cell=0), -1).expand_dims(
                    {"cell": list(missing_cells)}
                ),
            ],
            dim="cell",
        )

        out_inv.set_profiles(
            new_profile,
            indexes=new_indexes,
        )

    return out_inv



"""Utilities to use open street map data."""

import logging

import geopandas as gpd
from shapely.geometry import (
    GeometryCollection,
    LineString,
    MultiPolygon,
    Point,
    Polygon,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def osm_json_to_gdf(
    data: dict[str, list[dict[str, any]]],
    extract_tags: list[str] = [],
) -> gpd.GeoDataFrame:
    """Converts an OSM JSON to a GeoDataFrame.

    It proceeds the following way:
    - Nodes are converted to Points
    - Ways are converted to LineStrings or Polygons (if they are closed)
    - Relations are converted to MultiPolygons or GeometryCollections
    - Missing objects (Nodes, Ways, Relations) are ignored
    - objects without tags are ignored (simple geometry probably as ref for other objects)
    - tags are stored as strings in a tag column


    Example usage:

    ```python
    import requests

    url = "https://overpass-api.de/api/interpreter"
    bbox = (8.68, 49.40, 8.72, 49.42)
    query = f"[out:json];nwr({bbox[0]},{bbox[1]},{bbox[2]},{bbox[3]});out;"
    response = requests.get(url, params={"data": query})
    data = response.json()
    gdf = osm_json_to_gdf(data)

    ```

    :param data: OSM JSON data. This can be obtained by making a request to the OSM API.
    :return: A GeoDataFrame with the geometries and tags of the OSM data.

    """

    data_by_id = {x["id"]: x for x in data["elements"]}

    for tag in extract_tags:
        if tag in [
            "id",
            "geometry",
            "name",
            "type",
            "tags",
        ]:
            raise ValueError(f"Cannot extract tag {tag} as it is a reserved name")

    def geometry_from_nodes_ids(node_ids):
        nodes = [
            data_by_id[node_id]
            for node_id in node_ids
            # Remove nodes that are not in the data
            if node_id in data_by_id
        ]
        coords = [(node["lon"], node["lat"]) for node in nodes]

        if node_ids[0] == node_ids[-1]:
            return Polygon(coords)
        return LineString(coords)

    def resolve_way(element):
        """Ways are either polygons or linestrings."""
        return geometry_from_nodes_ids(element["nodes"])

    def resolve_relation(element):

        polys: list[Polygon] = []
        polys_roles: list[str] = []
        incomplete: list[LineString | Point] = []

        current_nodes: list[int] = []
        for member in element["members"]:

            if member["ref"] not in data_by_id:
                logger.debug(f"{member['ref']} not found")
                continue

            if member["type"] == "node":
                node = data_by_id[member["ref"]]
                incomplete.append(Point(node["lon"], node["lat"]))
                continue

            if member["type"] == "relation":
                incomplete.append(resolve_relation(data_by_id[member["ref"]]))
                continue

            assert (
                member["type"] == "way"
            ), f"Unexpected member type: {member['type']} in {element}"

            way = data_by_id[member["ref"]]

            if len(current_nodes) == 0:
                # Start the next polygon
                current_nodes.extend(way["nodes"])
                continue

            # Check if we can add the geom to the current polygon
            if way["nodes"][0] == current_nodes[-1]:
                current_nodes.extend(way["nodes"][1:])
            elif way["nodes"][-1] == current_nodes[0]:
                current_nodes = way["nodes"][:-1] + current_nodes
            elif way["nodes"][-1] == current_nodes[-1]:
                current_nodes.extend(way["nodes"][:-1:-1])
            elif way["nodes"][0] == current_nodes[0]:
                current_nodes = way["nodes"][1::-1] + current_nodes
            else:

                incomplete.append(geometry_from_nodes_ids(current_nodes))
                current_nodes = []
                continue

            # Check if we closed the polygon
            if current_nodes[0] == current_nodes[-1]:
                polys.append(geometry_from_nodes_ids(current_nodes))
                polys_roles.append(member["role"])
                current_nodes = []

        if len(current_nodes) > 0:
            incomplete.append(geometry_from_nodes_ids(current_nodes))

        # TODO: implement that we cut the inside of the polygons
        outer_polys = [p for p, r in zip(polys, polys_roles) if r == "outer"]
        inner_polys = [p for p, r in zip(polys, polys_roles) if r == "inner"]

        # Check if we return a multipolygon or a single polygon
        if len(incomplete) == 0:
            return MultiPolygon(polys)
        else:
            return GeometryCollection(polys + incomplete)

    def get_geometry_osm(element):
        """Resolves the geometry of an element."""
        match element["type"]:
            case "node":
                return Point(element["lon"], element["lat"])
            case "way":
                return resolve_way(element)
            case "relation":
                return resolve_relation(element)
            case _:
                raise ValueError(f"Unknown element type: {element['type']}")

    results_dict = [
        {
            "id": element["id"],
            # "geometry_type": element["tags"]["type"] if "type" in element["tags"] else None,
            "geometry": get_geometry_osm(element),
            "name": element["tags"].get("name", None),
            "type": element["type"],
            "tags": str(element["tags"]),
        }
        | {
            tag: element["tags"].get(tag, None)
            for tag in extract_tags
            if tag in element["tags"]
        }
        for element in data["elements"]
        if "tags" in element
    ]
    results_dict

    return gpd.GeoDataFrame(results_dict, crs="EPSG:4326")
